import logging
from abc import ABC, abstractmethod

import numpy as np

logger = logging.getLogger(__name__)

"""
Defines a generic Entity object, and an Entity convenience wrapper for creating Entities from numpy arrays.  
"""

DEFAULT_DTYPE = np.uint8


class Entity(ABC):
    """
    An Entity is a generalization of a synthetic object.  It could stand alone, or a composition of multiple entities.
    An Entity is composed of some data (represented by a numpy.ndarray), and an associated "valid" mask (numpy.ndarray).
    While the data can be a multi-channel, the mask must be a matrix of shape image.shape[0:2].  In other words,
    the mask applies to each channel of the data in the same way.  This is currently done for ease, but could be
    expanded such that a mask is specified for each channel of the image, if necessary.
    See the README for further details on how Entity objects are intended to be used in the TrojAI pipeline.
    """
    @abstractmethod
    def get_data(self) -> np.ndarray:
        """
        Get the data associated with the Entity
        :return: return a numpy.ndarray representing the image
        """
        pass

    @abstractmethod
    def get_mask(self) -> np.ndarray:
        """
        Get the mask associated with the Entity
        :return: return a numpy.ndarray representing the mask
        """
        pass


class GenericEntity(Entity):
    """
    A class which allows one to easily instantiate a generic Entity object with an image and associated mask
    """
    def __init__(self, data: np.ndarray, mask: np.ndarray = None) -> None:
        """
        Initialize the GenericEntity object, given an input image and associated mask
        :param data: The input image to be wrapped into an Entity
        :param mask: The associated mask to be wrapped into an Entity
        """
        self.pattern = data
        if mask is None:
            self.mask = np.ones(data.shape[0:2]).astype(bool)
        elif isinstance(mask, np.ndarray):
            if mask.shape[0:2] == self.pattern.shape[0:2]:
                self.mask = mask.astype(bool)
            else:
                msg = "Unknown Mask input - must be either None of a numpy.ndarray of same shape as arr_input"
                logger.error(msg)
                raise ValueError(msg)
        else:
            msg = "Unknown Mask input - must be either None of a numpy.ndarray of same shape as arr_input"
            logger.error(msg)
            raise ValueError(msg)

    def get_data(self) -> np.ndarray:
        """
        Get the data associated with the Entity
        :return: return a numpy.ndarray representing the image
        """
        return self.pattern

    def get_mask(self) -> np.ndarray:
        """
        Get the mask associated with the Entity
        :return: return a numpy.ndarray representing the mask
        """
        return self.mask



import logging
import os
from typing import Sequence

import cv2
import numpy as np
import pandas as pd
from numpy.random import RandomState
from sklearn.model_selection import train_test_split
from tqdm import tqdm
import math

import trojai.datagen.utils as utils
from .config import XFormMergePipelineConfig
from .constants import RANDOM_STATE_DRAW_LIMIT
from .entity import Entity, GenericEntity
from .merge import Merge
from .pipeline import Pipeline
from .transform import Transform

logger = logging.getLogger(__name__)

"""
Defines all functions and classes related to the transform+merge pipeline & data movement paradigm.
"""


def modify_clean_dataset(clean_dataset_rootdir: str, clean_csv_file: str,
                         output_rootdir: str, output_subdir: str, mod_cfg: XFormMergePipelineConfig,
                         method: str = 'insert', random_state_obj: RandomState = RandomState(1234)) -> None:
    """
    Modifies a clean dataset given a configuration

    :param clean_dataset_rootdir: root directory where the clean data lives
    :param clean_csv_file: filename of the CSV file which contains information about the clean data
                           The modification method determines which columns and information are expected
                           in the CSV file.
    :param output_rootdir: root directory where the modified data will be stored
    :param output_subdir: subdirectory where the modified data will be stored.  This is expected to be one level
                          below the root-directory, and can prove useful if different types of modifications are
                          stored in different subdirectories under the main root directory.  An example tree structure
                          might be:
                          root_data
                             - modification_1
                                 ... data ...
                             - modification_2
                                 ... data ...
    :param mod_cfg: A configuration object for creating a modified dataset
    :param method: Can be "insert" only/
                   In the insert method, the function takes the clean image, and inserts a specified Entity
                   (likely, a pattern) into the clean image.  Additional modes to be added!
    :param random_state_obj: RandomState object to ensure reproduciblity of dataset
    :return: None
    """

    try:
        os.makedirs(os.path.join(output_rootdir, output_subdir))
    except FileExistsError:
        pass

    # read in clean dataset
    clean_df = pd.read_csv(os.path.join(clean_dataset_rootdir, clean_csv_file))

    # identify which images will have triggers inserted into them
    random_state = random_state_obj.get_state()
    if mod_cfg.per_class_trigger_frac is not None:
        trigger_data, _ = train_test_split(clean_df,
                                           train_size=mod_cfg.per_class_trigger_frac,
                                           random_state=random_state_obj,
                                           stratify=clean_df['label'])
    else:
        trigger_data = clean_df
    # reset random state to be ensure reproduciblity regardless of # of splits
    random_state_obj.set_state(random_state)

    # generate the same # of triggers according to the configuration
    num_triggers = len(trigger_data)
    trigger_source_list = mod_cfg.trigger_list

    # run the xform function for each image & trigger combination
    for ii in tqdm(range(num_triggers), desc='Modifying Clean Dataset ...'):
        # select the trigger
        if trigger_source_list is not None and len(trigger_source_list) != 0:
            trigger = random_state_obj.choice(trigger_source_list, p=mod_cfg.trigger_sampling_prob)
        else:
            trigger = None
        img_random_state = RandomState(random_state_obj.randint(RANDOM_STATE_DRAW_LIMIT))

        if method.lower() == 'insert':
            fp = trigger_data.iloc[ii]['file']
            try:
                mask_fp = trigger_data.iloc[ii]['mask']
                mask = np.load(mask_fp)
            except KeyError:
                mask = None
            # load the background image
            bg = GenericEntity(cv2.imread(os.path.join(clean_dataset_rootdir, fp), cv2.IMREAD_UNCHANGED), mask)
            bg_xforms = mod_cfg.trigger_bg_xforms
            fg = trigger
            fg_xforms = mod_cfg.trigger_xforms
            merge_obj = mod_cfg.trigger_bg_merge
            postproc_xforms = mod_cfg.trigger_bg_merge_xforms
            # process data through the pipeline
            pipeline_obj = XFormMerge([[bg_xforms, fg_xforms]], [merge_obj], postproc_xforms)
            modified_img = pipeline_obj.process([bg, fg], img_random_state)
            logger.info("Inserted trigger=%s into image=%s" % (str(fg), str(bg)))
        else:
            msg = "Unknown/unimplemented data modification method!"
            logger.error(msg)
            raise ValueError(msg)

        output_fname = os.path.basename(trigger_data.iloc[ii]['file'])
        output_filename_fullpath = os.path.join(output_rootdir, output_subdir, output_fname)
        cv2.imwrite(output_filename_fullpath, modified_img.get_data())


class XFormMerge(Pipeline):
    """
    Implements a pipeline which is a series of cascading transform and merge
    operations.  The following diagram shows 4 objects as a series of serial
    transforms + merges.  Each pair of transformations is considered a
    "stage", and stages are processed in serial fashion.  In the diagram
    below, the data that each stage processes is:
        Stage1: obj1, obj2
        Stage2: Stage1_output, obj3
        Stage3: Stage2_output, obj4
    This extends in the obvious way to more objects, depending on how deep
    the pipeline is.

    obj1 --> xform  obj3 --> xform  obj4 --> xform
                  \               \               \
                   + --> xform --> + --> xform --> + --> xform output
                  /
    obj2 --> xform
    """
    def __init__(self, xform_list: Sequence[Sequence[Sequence[Transform]]], merge_list: Sequence[Merge],
                 final_xforms: Sequence[Transform] = None) -> None:
        """
        Create the pipeline object
        :param xform_list: Is a list of list of length 2, where each element
                           is a list of Transform objects.  For example:
                           [[Xform1_List, Xform2_List],
                            [Xform3_List, Xform4_List],
                            ...
                           ]
               Each stage of the Xform/Merge has a pair of transforms associated with it.  The first index into
               xform_list corresponds to the stage of the Xform/Merge, and the 2nd index corresponds to which object
               gets which transformation.
               [i][0] is used by the raw data
               [i][1] is used by the processed data from the previous stage,
                      except in the first stage, where the second raw image uses this transformation list
        :param merge_list: a list of Merge objects, where each index corresponds to the merge operation for that
                index's stage
        :param final_xforms: a list of final Transform objects
        """
        self.xform_list = xform_list
        self.merge_list = merge_list
        if final_xforms is None:
            self.final_xforms = []
        else:
            self.final_xforms = final_xforms

    @staticmethod
    def _process_two(bg: Entity, bg_xforms: Sequence[Transform], fg: Entity, fg_xforms: Sequence[Transform],
                     merge_obj: Merge, random_state_obj: RandomState) -> Entity:
        """
        Implements the following pipeline:
          bg --> xform
                       \
                        + --> output
                       /
          fg --> xform
        :param bg: Entity corresponding to "bg" in the diagram above
        :param bg_xforms: a sequence of transforms to be applied to the bg Entity
        :param fg: Entity corresponding to the "fg" in the diagram above
        :param fg_xforms: a sequence of transforms to be applied to the fg Entity
        :param merge_obj: a Merge object which corresponds to the "+" in the diagram above, and combines the two
                transformed objects
        :param random_state_obj: a random state to pass to the transforms and merge operation to ensure
                                 reproducibility of Entities produced by the pipeline
        :return: the Merged Entity according to the pipeline specification
        """

        if not isinstance(merge_obj, Merge):
            msg = "merge_obj argument must be of type: trojai.datagen.Merge"
            logger.error(msg)
            raise ValueError(msg)

        # perform some additional validation
        if bg is None and fg is None:
            msg = "Two None objects passing through the pipeline is an undefined operation!"
            logger.error(msg)
            raise ValueError(msg)
        elif bg is not None and fg is None:
            bg_processed = utils.process_xform_list(bg, bg_xforms, random_state_obj)
            logger.warning("Provided FG data is empty, only processing BG and returning without merge!")
            return bg_processed
        elif bg is None and fg is not None:
            fg_processed = utils.process_xform_list(fg, fg_xforms, random_state_obj)
            logger.warning("Provided BG data is empty, only processing FG and returning without merge!")
            return fg_processed
        else:
            # process the background & foreground images
            bg_processed = utils.process_xform_list(bg, bg_xforms, random_state_obj)
            fg_processed = utils.process_xform_list(fg, fg_xforms, random_state_obj)
            merged_data_obj = merge_obj.do(bg_processed, fg_processed, random_state_obj)
            logger.info("Processed BG and FG and merged!")
            return merged_data_obj

    def process(self, imglist: Sequence[Entity], random_state_obj: RandomState) -> Entity:
        """
        Processes the provided objects according to the Xform->Merge->Xform paradigm.
        :param imglist: a sequence of Entity objects to be processed according to the pipeline
        :param random_state_obj: a random state to pass to the transforms and merge operation to ensure
                                 reproducibility of Entities produced by the pipeline
        :return: the modified & combined Entity object
        """
        if len(imglist) < 2:
            raise ValueError("Need atleast 2 objects to process in a pipeline!")

        num_merges = len(imglist)-1
        num_expected_xforms = math.ceil(len(imglist)/2)
        if len(self.xform_list) != num_expected_xforms:
            msg = "Expected " + str(num_expected_xforms) + " xform(s) for " + str(num_expected_xforms) + " stage(s)!"
            logger.error(msg)
            raise ValueError(msg)
        if len(self.merge_list) != num_merges:
            msg = "Expected " + str(num_merges) + " merge object(s)!"
            logger.error(msg)
            raise ValueError(msg)
        for xl in self.xform_list:
            if len(xl) != 2:
                msg = "Expected 2 xforms per merge operation!"
                logger.error(msg)
                raise ValueError(msg)

        # process the data through the pipeline
        z = None
        for imglist_idx in range(1, len(imglist)):
            mergeobj_idx = imglist_idx-1
            if imglist_idx == 1:
                merge_input1 = imglist[0]
                merge_input2 = imglist[imglist_idx]
            else:
                merge_input1 = imglist[imglist_idx]
                merge_input2 = z
            merge_input1_xforms = self.xform_list[mergeobj_idx][0]
            merge_input2_xforms = self.xform_list[mergeobj_idx][1]
            z = XFormMerge._process_two(merge_input1, merge_input1_xforms, merge_input2,
                                        merge_input2_xforms, self.merge_list[mergeobj_idx], random_state_obj)

        logger.info("XFormMerged input images")
        # process the final xform
        z_final = utils.process_xform_list(z, self.final_xforms, random_state_obj)
        return z_final



from abc import ABC, abstractmethod
from typing import Union, Sequence
import logging

from .config import ModelGeneratorConfig

logger = logging.getLogger(__name__)


class ModelGeneratorInterface(ABC):
    """Generates models based on requested data and saves each to a file."""
    def __init__(self, configs: Union[ModelGeneratorConfig, Sequence[ModelGeneratorConfig]]):
        """
        :param configs: configuration objects that specify how to generate models for a single experiment
        """
        self.configs = configs
        if not isinstance(self.configs, Sequence):
            self.configs = [self.configs]

    @abstractmethod
    def run(self) -> None:
        """
        Train and save models as specified.
        :return: None
        """
        pass


def validate_model_generator_interface_input(configs: Union[ModelGeneratorConfig, Sequence[ModelGeneratorConfig]]) \
        -> None:
    """
    Validates a ModelGeneratorConfig
    :param configs: (ModelGeneratorConfig or sequence) configurations to be used for model generation
    :return None
    """
    if not (isinstance(configs, ModelGeneratorConfig) or isinstance(configs, Sequence)):
        err_msg = "Expected a ModelGeneratorConfig object or sequence of ModelGeneratorConfig objects for " \
                  "argument 'configs', instead got type: {}".format(type(configs))
        logger.error(err_msg)
        raise TypeError(err_msg)
    if isinstance(configs, Sequence) and len(configs) == 0:
        err_msg = "Emtpy sequence provided for 'configs' argument."
        logger.error(err_msg)
        raise RuntimeError(err_msg)
    if isinstance(configs, Sequence):
        for cfg in configs:
            if not isinstance(cfg, ModelGeneratorConfig):
                err_msg = "non-'ModelGeneratorConfig' type included in argument 'configs': {}".format(type(cfg))
                logger.error(err_msg)
                raise TypeError(err_msg)
    logger.debug("Configuration validated successfully!")



import logging
from typing import Union, Sequence
import torch

from tqdm import tqdm

from .model_generator_interface import ModelGeneratorInterface, validate_model_generator_interface_input
from .config import ModelGeneratorConfig
from .runner import Runner
from .config import modelgen_cfg_to_runner_cfg

logger = logging.getLogger(__name__)


class ModelGenerator(ModelGeneratorInterface):
    """Generates models based on requested data and saves each to a file."""
    def __init__(self, configs: Union[ModelGeneratorConfig, Sequence[ModelGeneratorConfig]], *args, **kwargs):
        """
        :param configs: (ModelGeneratorConfig or sequence) ModelGeneratorConfig objects configured to generate models
            for a single experiment
        """
        super().__init__(configs)
        self.validate()

    def run(self, *args, **kwargs) -> None:
        """
        Train and save models as specified.
        :return: None
        """
        loop = tqdm(self.configs, desc='Configurations')
        for cfg in loop:
            loop.set_postfix_str(cfg.experiment_cfg['name'])
            for i in range(cfg.num_models):
                filename = None
                run_id = None
                if cfg.filenames is not None:
                    if isinstance(cfg.filenames, str):
                        filename = cfg.filenames
                    else:
                        filename = cfg.filenames[i]
                elif cfg.run_ids is not None:
                    run_id = cfg.run_ids[i]

                run_cfg = modelgen_cfg_to_runner_cfg(cfg, run_id=run_id, filename=filename)
                runner = Runner(run_cfg, persist_metadata=cfg.experiment_cfg)
                runner.run()

                # clear up memory between runs
                torch.cuda.empty_cache()

    def validate(self) -> None:
        """
        Validate the provided input when constructing the ModelGenerator interface
        """
        validate_model_generator_interface_input(self.configs)



from abc import ABC, abstractmethod

from numpy.random import RandomState

from .entity import Entity

"""
Defines a generic Merge object.  
"""


class Merge(ABC):
    """
    A Merge is defined as an operation on two Entities and returns a single Entity
    """
    @abstractmethod
    def do(self, obj1: Entity, obj2: Entity, random_state_obj: RandomState) -> Entity:
        """
        Perform the actual merge operation
        :param obj1: the first Entity to be merged
        :param obj2: the second Entity to be merged
        :param random_state_obj: a numpy.random.RandomState object to ensure reproducibility
        :return: the merged Entity
        """
        pass



from abc import ABC, abstractmethod

from numpy.random import RandomState

from .entity import Entity

"""
Defines a generic Transform object.  
"""


class Transform(ABC):
    """
    A Transform is defined as an operation on an Entity.
    """
    @abstractmethod
    def do(self, input_obj: Entity, random_state_obj: RandomState) -> Entity:
        """
        Perform the specified transformation
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: a random state used to maintain reproducibility through transformations
        :return: the transformed Entity
        """
        pass



import logging
from typing import Iterable

import copy
from numpy.random import RandomState

from .entity import Entity
from .transform import Transform

logger = logging.getLogger(__name__)

"""
Contains general utilities helpful for data generation
"""


def process_xform_list(input_obj: Entity, xforms: Iterable[Transform], random_state_obj: RandomState) -> Entity:
    """
    Processes a list of transformations in a serial fashion on a copy of the input X
    :param input_obj: input object which should be transformed by the list of
              transformations
    :param xforms: a list of Transform objects
    :param random_state_obj:
    :return: The transformed object
    """
    input_obj_copy = copy.deepcopy(input_obj)
    for xform in xforms:
        logger.info("Applying:%s to input_obj: %s" % (str(xform), str(input_obj_copy)))
        input_obj_copy = xform.do(input_obj_copy, random_state_obj)
    return input_obj_copy



from typing import Union, Tuple
import glob
import logging
import os

import numpy as np
import pandas as pd
from numpy.random import RandomState
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from .label_behavior import LabelBehavior

logger = logging.getLogger(__name__)

"""
Module which contains functionality for generating experiments
"""


class ClassicExperiment:
    """
    Defines a classic experiment, which consists of: 1) a specification of the clean data 2) a specification of the
    modified (triggered) data, and 3) a specification of the split of triggered/clean data for training/testing
    the model
    """
    def __init__(self, data_root_dir: str, trigger_label_xform: LabelBehavior, stratify_split: bool = True) -> None:
        """
        Initializes a Classic experiment object
        :param data_root_dir: the root directory under which all data lives under.  The expected directory structure
                for any dataset is as follows:
                root_dir
                  |- clean_data
                  |- modification_1
                  |- modification_2
                  |- ...
                This is needed so that the proper relative path can be computed from the root directory.
                Additionally, it is required that filenames correspond across the different subfolders under
                root_dir. Practically, this means
        :param trigger_label_xform: a LabelBehavior object specifying how triggered data is changed
        :param stratify_split: if True, then data is split such that each class has the same number of samples in
                the produced experiment
        """
        self.data_root_dir = data_root_dir
        self.stratify_split = stratify_split
        self.trigger_label_xform = trigger_label_xform

    def create_experiment(self, clean_data_csv: str, experiment_data_folder: str,
                          mod_filename_filter: str = '*', split_clean_trigger: bool = False,
                          trigger_frac: float = 0.2, random_state_obj: RandomState = RandomState(1234)) \
            -> Union[Tuple, pd.DataFrame]:
        """
        Creates an "experiment," which is a dataframe defining the data that should be used, and whether that data is
         triggered or not, and the true & actual label associated with that data point.
        TODO:
          [] - Have ability to accept multiple mod_data_folders such that we can sample from them all at a specified
               probability to have different triggers
        :param clean_data_csv: path to file which contains a CSV specification of the clean data. The CSV file is
                expected to have the following columns: [file, label]
        :param experiment_data_folder: the folder which contains the data to mix with for the experiment.
        :param mod_filename_filter: a string filter for determining which files in the folder to consider, if only a
                a subset is to be considered for sampling
        :param split_clean_trigger: if True, then we return a list of DataFrames, where the triggered & non-triggered
                data are combined into one DataFrame, if False, we concatenate the triggered and non-triggered data
                into one DataFrame
        :param trigger_frac: the fraction of data which which should be triggered
        :param random_state_obj: random state object
        :return: a dataframe of the data which consists of the experiment.  The DataFrame has the following columns:
                    file, true_label, train_label, triggered
                    file - the file path of the data
                    true_label - the actual label of the data
                    train_label - the label of the data the model should be trained on.
                                  This will be equal to true_label *if* triggered==False
                    triggered -  a boolean value indicating whether this particular sample has a Trigger or not
        """
        logger.info("Creating experiment from clean_data:%s modified_data:%s" %
                    (clean_data_csv, experiment_data_folder))
        # get absolute paths to avoid ambiguities when generating output paths
        experiment_data_folder = os.path.abspath(experiment_data_folder)

        clean_df = pd.read_csv(clean_data_csv)
        clean_df['filename_only'] = clean_df['file'].map(os.path.basename)
        # find list of files in the mod data folder that match the input filter
        num_trigger = int(len(clean_df) * trigger_frac)
        mod_flist = glob.glob(os.path.join(experiment_data_folder, mod_filename_filter))
        if not self.stratify_split:
            mod_flist_subset = random_state_obj.choice(mod_flist, num_trigger, replace=False)
            logger.info("Created unstratified dataset from %s for including in experiment" % (experiment_data_folder,))
        else:
            # get overlap between files which exist in the directory and files which were converted
            # and pick stratification based on the original label
            orig_flist = set(clean_df['filename_only'])
            mod_flist_fname_only = set([os.path.basename(x) for x in mod_flist])
            common_flist = list(orig_flist.intersection(mod_flist_fname_only))
            df_subset_to_stratify = clean_df[clean_df['filename_only'].isin(common_flist)]
            # get the trigger fraction percentage based on class-label stratification
            if trigger_frac > 0:
                try:
                    df_flist, _ = train_test_split(df_subset_to_stratify,
                                                   train_size=num_trigger,
                                                   random_state=random_state_obj,
                                                   stratify=df_subset_to_stratify['label'])
                    logger.info("Created stratified dataset from %s for including in experiment" %
                                (experiment_data_folder,))
                except ValueError as e:
                    logger.exception(e)
                    logger.error("Error creating experiment, likely because the fraction of triggered data specified "
                                 "creates a data split where not all classes are represented!")
                    raise ValueError(e)
            else:
                # empty dataframe with no entries, meaning that no data is triggered
                df_flist = pd.DataFrame(columns=['file', 'label', 'filename_only'])
                logger.info("Using all data points in %s for experiment" % (experiment_data_folder,))
            mod_flist_subset = list(df_flist['filename_only'].map(lambda x: os.path.join(experiment_data_folder, x)))

        # compose into an experiment CSV file
        clean_df.rename(columns={'file': 'file',
                                 'label': 'true_label',
                                 'filename_only': 'filename_only'},
                        inplace=True)
        clean_df['train_label'] = clean_df['true_label']
        clean_df['triggered'] = False
        # change filename to be relative to root-folder rather than subfolder
        clean_data_folder = os.path.dirname(clean_data_csv)
        clean_data_rootfolder_relpath = os.path.relpath(clean_data_folder, self.data_root_dir)
        clean_df['file'] = clean_df['file'].map(lambda x: os.path.join(clean_data_rootfolder_relpath, x))

        # create a dataframe of the triggered data
        num_mod = len(mod_flist_subset)
        mod_files_true_labels = np.empty(num_mod, dtype=clean_df['train_label'].dtype)
        mod_files_triggered_labels = np.empty(num_mod, dtype=clean_df['train_label'].dtype)
        for ii, f in enumerate(tqdm(mod_flist_subset)):
            fname_only = os.path.basename(f)
            # search for the filename in the original data to get the true label associated with this file
            clean_data_assoc_label_series = clean_df[clean_df['filename_only'] == fname_only]['true_label']
            if len(clean_data_assoc_label_series) > 1:
                raise ValueError("Multiple filenames match - duplication detected for " + str(fname_only) + "!")
            if len(clean_data_assoc_label_series) == 0:
                raise ValueError("File:" + str(f) + " seems to have disappeared!")
            clean_data_assoc_label = clean_data_assoc_label_series.iat[0]
            mod_files_true_labels[ii] = clean_data_assoc_label
            # modify the label behavior according to the specified behavior
            mod_files_triggered_labels[ii] = self.trigger_label_xform.do(clean_data_assoc_label)
        clean_df.drop(['filename_only'], axis=1, inplace=True)

        triggered_df = pd.DataFrame(mod_flist_subset, columns=['file'])
        # adjust the paths to the filename so that it is relative to the data root directory
        mod_data_rootfolder_relpath = os.path.relpath(experiment_data_folder, self.data_root_dir)
        triggered_df['file'] = triggered_df['file'].map(
            lambda x: os.path.join(mod_data_rootfolder_relpath, os.path.basename(x)))
        triggered_df['true_label'] = mod_files_true_labels
        triggered_df['train_label'] = mod_files_triggered_labels
        triggered_df['triggered'] = True

        # now-subsample the clean-df for the percentage of data we want
        clean_df_subset = clean_df.sample(frac=1 - trigger_frac, random_state=random_state_obj)
        if split_clean_trigger:
            return clean_df_subset, triggered_df
        else:
            # merge the dataframes
            # TODO: decide whether ignore_index=True is the correct behavior here! My initial
            #   opinion is that it doesn't matter, since likely the dataframe will be written
            #   to a CSV with index=None, but the action item is to think through the potential
            #   other usecases
            return pd.concat([clean_df_subset, triggered_df])



import collections
import logging
import os
from typing import Sequence, Any
import copy
import cloudpickle as pickle
import numpy as np
from tqdm import tqdm

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.tensorboard import SummaryWriter
from torchtext.data.iterator import Iterator as TextDataIterator
from torchtext.data.iterator import BucketIterator
import torchtext

from .datasets import CSVTextDataset
from .training_statistics import BatchStatistics, EpochStatistics
from .optimizer_interface import OptimizerInterface
from .default_optimizer import _eval_binary_acc
from .config import LSTMOptimizerConfig
from .constants import VALID_OPTIMIZERS

logger = logging.getLogger(__name__)


class LSTMOptimizer(OptimizerInterface):
    """
    An optimizer for training and testing LSTM models. Currently in a prototype state.
    """

    def __init__(self, optimizer_cfg: LSTMOptimizerConfig = None):
        """
        Initializes the optimizer with an LSTMOptimizerConfig
        :param optimizer_cfg: the configuration used to initialize the LSTMOptimizer
        """
        if optimizer_cfg is None:
            logger.info("Using default parameters to setup Optimizer!")
            self.optimizer_cfg = LSTMOptimizerConfig()
        elif not isinstance(optimizer_cfg, LSTMOptimizerConfig):
            msg = "optimizer_cfg must be of type LSTMOptimizerConfig"
            logger.error(msg)
            raise TypeError(msg)
        else:
            self.optimizer_cfg = optimizer_cfg

        # setup parameters for training here
        self.device = self.optimizer_cfg.training_cfg.device

        self.loss_function_str = self.optimizer_cfg.training_cfg.objective.lower()
        if self.loss_function_str == "cross_entropy_loss":
            self.loss_function = nn.CrossEntropyLoss()
        elif self.loss_function_str == 'bcewithlogitsloss':
            self.loss_function = nn.BCEWithLogitsLoss()
        self.loss_function.to(self.device)

        self.lr = self.optimizer_cfg.training_cfg.lr
        self.optimizer_str = self.optimizer_cfg.training_cfg.optim.lower()
        self.optimizer = None

        self.batch_size = self.optimizer_cfg.training_cfg.batch_size
        self.num_epochs = self.optimizer_cfg.training_cfg.epochs

        self.str_description = "{'batch_size':%d, 'num_epochs':%d, 'device':'%s', 'lr':%.5e, 'loss_function':'%s', " \
                               "'optimizer':'%s'}" % \
                               (self.batch_size, self.num_epochs, self.device.type, self.lr, self.loss_function_str,
                                self.optimizer_str)

        # setup configuration for logging and tensorboard
        self.num_batches_per_logmsg = self.optimizer_cfg.reporting_cfg.num_batches_per_logmsg
        self.num_epochs_per_metrics = self.optimizer_cfg.reporting_cfg.num_epochs_per_metrics
        self.num_batches_per_metrics = self.optimizer_cfg.reporting_cfg.num_batches_per_metrics
        # NOTE: the configuration parameter 'num_batches_per_val_dataset_metrics' has the most significant impact on
        #  training performance.  It can be useful for debugging development of a model, but when scaling up,
        #  set this parameter to None to disable computing the validation dataset and speed up training
        self.num_batches_per_val_dataset_metrics = self.optimizer_cfg.reporting_cfg.num_batches_per_val_dataset_metrics
        if self.device.type == 'cpu' and self.num_batches_per_val_dataset_metrics is not None:
            logger.warning('Training will be VERY SLOW on a CPU with num_batches_per_val_dataset_metrics set to a '
                           'value other than None.  If validation dataset metrics are still desired, '
                           'consider increasing this value to speed up training')

        tensorboard_output_dir = self.optimizer_cfg.reporting_cfg.tensorboard_output_dir
        self.tb_writer = SummaryWriter(tensorboard_output_dir)

        optimizer_cfg_str = 'Optimizer[%s] Configured as: loss[%s], learning-rate[%.5e], batch-size[%d] ' \
                            'num-epochs[%d] Device[%s]' % \
                            (self.optimizer_str, str(self.loss_function), self.lr, self.batch_size, self.num_epochs,
                             self.device.type)
        reporting_cfg_str = 'Reporting Configured as: num_batches_per_log_message[%d] tensorboard_dir[%s]' % \
                            (self.num_batches_per_logmsg, tensorboard_output_dir)
        metrics_capture_str = 'Metrics capturing configured as: num_epochs_per_metric[%d] ' \
                              'num_batches_per_epoch_per_metric[%d]' % \
                              (self.num_epochs_per_metrics, self.num_batches_per_metrics)
        logger.info(self.str_description)
        logger.info(optimizer_cfg_str)
        logger.info(reporting_cfg_str)
        logger.info(metrics_capture_str)

    def __str__(self):
        return self.str_description

    def __deepcopy__(self, memodict={}):
        optimizer_cfg_copy = copy.deepcopy(self.optimizer_cfg)
        # WARNING: this assumes that none of the derived attributes have been changed after construction!
        return LSTMOptimizer(LSTMOptimizerConfig(optimizer_cfg_copy.training_cfg,
                                                 optimizer_cfg_copy.reporting_cfg))

    def __eq__(self, other: Any):
        try:
            if self.optimizer_cfg == other.optimizer_cfg:
                # we still check the derived attributes to ensure that they remained the same after
                # after construction
                if self.device.type == other.device.type and self.loss_function_str == other.loss_function_str and \
                        self.lr == other.lr and self.optimizer_str == other.optimizer_str and \
                        self.batch_size == other.batch_size and self.num_epochs == other.num_epochs and \
                        self.str_description == other.str_description and \
                        self.num_batches_per_logmsg == other.num_batches_per_logmsg and \
                        self.num_epochs_per_metrics == other.num_epochs_per_metrics and \
                        self.num_batches_per_metrics == other.num_batches_per_metrics and \
                        self.num_batches_per_val_dataset_metrics == other.num_batches_per_val_dataset_metrics and \
                        self.tb_writer.log_dir == other.tb_writer.log_dir:
                    return True
            else:
                return False
        except AttributeError:
            return False

    def get_cfg_as_dict(self) -> dict:
        output_dict = dict(device=str(self.device.type),
                           epochs=self.num_epochs,
                           batch_size=self.batch_size,
                           learning_rate=self.lr,
                           optim=self.optimizer_str,
                           objective=self.loss_function_str)
        return output_dict

    def get_device_type(self) -> str:
        """
        :return: a string representing the device used to train the model
        """
        return self.device.type

    def save(self, fname: str) -> None:
        """
        Saves the configuration object used to construct the LSTMOptimizer.
        NOTE: because the LSTMOptimizer object itself is not persisted, but rather the
          LSTMOptimizerConfig object, the state of the object does not persist!
        :param fname: the filename to save the LSTMOptimizer's configuration.
        """
        self.optimizer_cfg.save(fname)

    @staticmethod
    def load(fname: str) -> OptimizerInterface:
        """
        Reconstructs an LSTMOptimizer, by loading the configuration used to construct the original
        LSTMOptimizer, and then creating a new LSTMOptimizer object from the saved configuration
        :param fname: The filename of the saved LSTMOptimizer
        :return: an LSTMOptimizer object
        """
        with open(fname, 'rb') as f:
            loaded_optimzier_cfg = pickle.load(f)
        return LSTMOptimizer(loaded_optimzier_cfg)

    def _eval_loss_function(self, y_hat: torch.Tensor, y_truth: torch.Tensor) -> torch.Tensor:
        """
        Wrapper for evaluating the loss function to abstract out any data casting we need to do
        :param y_hat: the predicted y-value
        :param y_truth: the actual y-value
        :return: the loss associated w/ the prediction and actual
        """
        if self.loss_function_str == "cross_entropy_loss":
            train_loss = self.loss_function(y_hat, y_truth.long())
        else:
            train_loss = self.loss_function(y_hat, y_truth)
        return train_loss

    @staticmethod
    def train_val_dataset_split(dataset: torchtext.data.Dataset, split_amt: float) \
            -> (torchtext.data.Dataset, torchtext.data.Dataset):
        """
        Splits a torchtext dataset (of type: torchtext.data.Dataset) into train/test.
        NOTE: although this has the same functionality as default_optimizer.train_val_dataset_split, it works with a
         torchtext.data.Dataset object rather than torch.utils.data.Dataset.
        TODO:
          [ ] - specify random seed to torch splitter
        :param dataset: the dataset to be split
        :param split_amt: fraction specificing the validation dataset size relative to the whole.  1-split_amt will
                          be the size of the training dataset
        :return: a tuple of the train and validation datasets
        """

        if split_amt < 0 or split_amt > 1:
            msg = "Dataset split amount must be between 0 and 1"
            logger.error(msg)
            raise ValueError(msg)

        train_dataset, val_dataset = dataset.split(1 - split_amt)
        return train_dataset, val_dataset

    def convert_dataset_to_dataiterator(self, dataset: CSVTextDataset) -> TextDataIterator:
        # NOTE: We compare types in this manner, even though it is considered a Python anti-pattern.
        #  https://docs.quantifiedcode.com/python-anti-patterns/readability/do_not_compare_types_use_isinstance.html
        #  It seems more efficient to store the datatype, and pass it around, rather than the dataset itself.  We
        #  can't check the type of the dataset argument directly b/c if it is split, it may change Datatypes,
        #  as is the case when the random_split function is called.

        # NOTE: shuffle argument is not used here b/c it is shuffled on the input, but is it better to do the
        #  shuffling here (or another place?)
        return BucketIterator(dataset, self.batch_size, device=self.device, sort_within_batch=True)

    def train(self, model: torch.nn.Module, dataset: CSVTextDataset, train_val_split: float = 0.0) \
            -> (torch.nn.Module, Sequence[EpochStatistics]):
        """
        Train the network.
        :param model: the model to train
        :param dataset: the dataset to train the network on
        :param train_val_split: the % of training data to use as validation data
        :return: the trained network, list of EpochStatistics objects
        """
        model = model.to(self.device)

        model.train()  # put network into training mode
        if self.optimizer_str == 'adam':
            self.optimizer = optim.Adam(model.parameters(), lr=self.lr)
        elif self.optimizer_str not in VALID_OPTIMIZERS:
            msg = self.optimizer_str + " is not a supported optimizer!"
            logger.error(msg)
            raise ValueError(msg)
        else:
            msg = self.optimizer_str + " not yet implemented!"
            logger.error(msg)
            raise NotImplementedError(msg)

        # split into train & validation datasets, and setup data loaders according to their type
        train_dataset, val_dataset = LSTMOptimizer.train_val_dataset_split(dataset, train_val_split)
        train_loader = self.convert_dataset_to_dataiterator(train_dataset)
        val_loader = self.convert_dataset_to_dataiterator(val_dataset)

        # before training - we should transfer the embedding from the learned to the model weights
        pretrained_embeddings = dataset.text_field.vocab.vectors
        model.embedding.weight.data.copy_(pretrained_embeddings)
        UNK_IDX = dataset.text_field.vocab.stoi[dataset.text_field.unk_token]
        PAD_IDX = dataset.text_field.vocab.stoi[dataset.text_field.pad_token]
        model.embedding.weight.data[UNK_IDX] = torch.zeros(model.embedding_dim)
        model.embedding.weight.data[PAD_IDX] = torch.zeros(model.embedding_dim)

        # use validation in training? provide as option?
        all_epochs_stats = []
        for epoch_idx, epoch in enumerate(range(self.num_epochs)):
            compute_batch_stats = True if epoch_idx % self.num_epochs_per_metrics == 0 else False
            batches_stats = self.train_epoch(model, train_loader, val_loader, epoch, compute_batch_stats)

            if compute_batch_stats:
                epoch_training_stats = EpochStatistics(epoch_idx)
                epoch_training_stats.add_batch(batches_stats)
                all_epochs_stats.append(epoch_training_stats)

        return model, all_epochs_stats

    def train_epoch(self, model: nn.Module, train_loader: TextDataIterator, val_loader: TextDataIterator,
                    epoch_num: int, compute_batch_stats: bool = True,
                    avg_loss_num_batches: int = 100):
        """
        Runs one epoch of training on the specified model
        TODO:
          [ ] - pass in the avg_loss_num_batches parameter cleanly and configurably
        :param model: the model to train for one epoch
        :param train_loader: a DataLoader object pointing to the training dataset
        :param val_loader: a DataLoader object pointing to the validation dataset
        :param epoch_num: the epoch number that is being trained
        :param compute_batch_stats: if True, computes statistics for the batch based on the reporting configuration
                specified in the initialization of the optimizer
        :param avg_loss_num_batches: the number of batches of data to accumulate to compute average loss
        :return: a list of statistics for batches where statistics were computed
        """

        pid = os.getpid()
        train_dataset_len = len(train_loader.dataset)
        train_loader_len = len(train_loader)
        loop = tqdm(train_loader)

        # NOTE: potential speed-up by not computing the average ... but this seems like premature optimization to me
        avg_train_loss_circbuf = collections.deque(maxlen=avg_loss_num_batches)
        avg_val_loss_vec = np.empty(len(val_loader.dataset))

        train_n_correct, train_n_total = 0, 0
        val_n_correct, val_n_total = 0, 0
        val_acc, val_loss = None, None
        batch_stats = []
        for batch_idx, batch in enumerate(loop):
            # put network into training mode & zero out previous gradient computations
            model.train()
            self.optimizer.zero_grad()

            # get predictions based on input & weights learned so far
            text, text_lengths = batch.text
            predictions = model(text, text_lengths).squeeze(1)

            # compute metrics
            batch_train_loss = self._eval_loss_function(predictions, batch.label)
            running_train_acc, train_n_total, train_n_correct = _eval_binary_acc(predictions, batch.label,
                                                                                 n_total=train_n_total,
                                                                                 n_correct=train_n_correct)

            # compute gradient
            batch_train_loss.backward()
            self.optimizer.step()

            if len(val_loader) > 0 and (self.num_batches_per_val_dataset_metrics is not None) and \
                    ((batch_idx % self.num_batches_per_val_dataset_metrics == 0) or
                     (batch_idx % self.num_batches_per_metrics == 0)):
                # last condition ensures metrics are computed for storage put model into evaluation mode
                model.eval()
                # turn off auto-grad for validation set computation
                with torch.no_grad():
                    for val_batch_idx, batch in enumerate(val_loader):
                        text, text_lengths = batch.text
                        predictions = model(text, text_lengths).squeeze(1)

                        val_loss_tensor = self._eval_loss_function(predictions, batch.label)
                        val_loss = val_loss_tensor.item()
                        val_acc, val_n_total, val_n_correct = _eval_binary_acc(predictions, batch.label,
                                                                               n_total=val_n_total,
                                                                               n_correct=val_n_correct)
                        avg_val_loss_vec[val_batch_idx] = val_loss

                avg_val_loss = np.mean(avg_val_loss_vec)
                self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                          '-avg_validation_loss', avg_val_loss)

            # NOTE: should these be options to compute training accuracy, etc, as well?
            avg_train_loss_circbuf.append(batch_train_loss.item())
            avg_train_loss = np.mean(avg_train_loss_circbuf)

            loop.set_description('Epoch {}/{}'.format(epoch_num + 1, self.num_epochs))
            loop.set_postfix(avg_train_loss=avg_train_loss)

            # report batch statistics to tensorflow
            self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-train_loss',
                                      batch_train_loss.item())
            self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-avg_train_loss',
                                      avg_train_loss)
            self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-running_train_acc',
                                      running_train_acc)
            if len(val_loader) > 0 and self.num_batches_per_val_dataset_metrics is not None:
                self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-val_acc', val_acc)

            # save batch statistics
            if compute_batch_stats and (batch_idx % self.num_batches_per_metrics == 0):
                batch_stat = BatchStatistics(batch_idx, running_train_acc, batch_train_loss.item(),
                                             val_acc, val_loss)
                batch_stats.append(batch_stat)

            if batch_idx % self.num_batches_per_logmsg == 0:
                logger.info('{}\tTrain Epoch: {} [{}/{} ({:.0f}%)]\tAvgTrainLoss: {:.6f}\tTrainAcc: {:.6f}'.format(
                    pid, epoch_num, batch_idx * len(batch), train_dataset_len,
                                    100. * batch_idx / train_loader_len, avg_train_loss, running_train_acc))

        return batch_stats

    def test(self, model: nn.Module, clean_data: CSVTextDataset, triggered_data: CSVTextDataset) -> dict:
        """
        Test the trained network
        :param model: the trained module to run the test data through
        :param clean_data: the clean Dataset
        :param triggered_data: the triggered Dataset, if None, not computed
        :return: a dictionary of the statistics on the clean and triggered data (if applicable)
        """
        test_data_statistics = {}
        model.eval()

        data_loader = self.convert_dataset_to_dataiterator(clean_data)
        loop = tqdm(data_loader)

        # test type is classification accuracy on clean and triggered data
        test_n_correct = 0
        test_n_total = 0
        with torch.no_grad():
            for batch_idx, batch in enumerate(loop):
                text, text_lengths = batch.text
                predictions = model(text, text_lengths).squeeze(1)
                test_acc, test_n_total, test_n_correct = _eval_binary_acc(predictions, batch.label,
                                                                               n_total=test_n_total,
                                                                               n_correct=test_n_correct)
        test_data_statistics['clean_accuracy'] = test_acc
        test_data_statistics['clean_n_total'] = test_n_total
        logger.info("Accuracy on clean test data: %0.02f" %
                    (test_data_statistics['clean_accuracy'],))

        if triggered_data is None:
            return test_data_statistics

        data_loader = self.convert_dataset_to_dataiterator(triggered_data)
        test_n_correct = 0
        test_n_total = 0
        with torch.no_grad():
            for batch_idx, batch in enumerate(data_loader):
                text, text_lengths = batch.text
                predictions = model(text, text_lengths).squeeze(1)
                test_acc, test_n_total, test_n_correct = _eval_binary_acc(predictions, batch.label,
                                                                               n_total=test_n_total,
                                                                               n_correct=test_n_correct)
        test_data_statistics['triggered_accuracy'] = test_acc
        test_data_statistics['triggered_n_total'] = test_n_total
        logger.info("Accuracy on triggered test data: %0.02f" %
                    (test_data_statistics['triggered_accuracy'],))
        return test_data_statistics



import logging
import os
from typing import Callable, Union

import cv2
import pandas as pd
import torch
import torchtext.data
from numpy.random import RandomState
from torch.utils.data import Dataset
from tqdm import tqdm

logger = logging.getLogger(__name__)

"""
Defines various types of datasets that are used by the DataManager
"""


class CSVDataset(Dataset):
    """
    Defines a dataset that is represented by a CSV file with columns "file", "train_label", and optionally
    "true_label". The file column should contain the path to the file that contains the actual data,
    and "train_label" refers to the label with which the data should be trained.  "true_label" refers to the actual
    label of the data point, and can differ from train_label if the dataset is poisoned.  A CSVDataset can support
    any underlying data that can be loaded on the fly and fed into the model (for example: image data)
    """
    def __init__(self, path_to_data: str, csv_filename:str , true_label=False, path_to_csv=None, shuffle=False,
                 random_state: Union[int, RandomState]=None,
                 data_loader: Union[str, Callable] = 'default_image_loader',
                 data_transform=lambda x: x, label_transform=lambda l: l):
        """
        Initializes a CSVDataset object.
        :param path_to_data: the root folder where the data lives
        :param csv_filename: the CSV file specifying the actual data points
        :param true_label (bool): if True, then use the column "true_label" as the label associated with each
        datapoint.  If False (default), use the column "train_label" as the label associated with each datapoint
        :param path_to_csv: If not None, specifies the folder where the CSV file lives.  If None, it is assumed that
            the CSV file lives in the same directory as the path_to_data
        :param shuffle: if True, the dataset is shuffled before loading into the model
        :param random_state: if specified, seeds the random sampler when shuffling the data
        :param data_loader: either a string value (currently only supports `default_image_loader`), or a callable
            function which takes a string input of the file path and returns the data
        :param data_transform: a callable function which is applied to every data point before it is fed into the
            model. By default, this is an identity operation
        :param label_transform: a callable function which is applied to every label before it is fed into the model.
            By default, this is an identity operation.
        """
        self.path_to_data = path_to_data
        if path_to_csv is None:
            path_to_csv = path_to_data
        else:
            path_to_csv = path_to_csv
        self.label = 'train_label'
        if true_label:
            self.label = 'true_label'
        self.data_df = pd.read_csv(os.path.join(path_to_csv, csv_filename))
        if shuffle:
            self.data_df = self.data_df.sample(frac=1, random_state=random_state).reset_index(drop=True)
        if not callable(data_loader):
            if data_loader == 'default_image_loader':
                self.data_loader = lambda img_loc: torch.from_numpy(cv2.imread(img_loc, cv2.IMREAD_UNCHANGED)).float()
            else:
                msg = "Unknown data loader specified!"
                logger.error(msg)
                raise ValueError(msg)
        else:
            self.data_loader = data_loader
        self.data_transform = data_transform
        self.label_transform = label_transform

    def __getitem__(self, item):
        data_loc = os.path.join(self.path_to_data, self.data_df.iloc[item]["file"])
        data = self.data_loader(data_loc)
        data = self.data_transform(data)
        label = self.data_df.iloc[item][self.label]
        label = self.label_transform(label)
        return data, label

    def __len__(self):
        return len(self.data_df)


class CSVTextDataset(torchtext.data.Dataset):
    """
    Defines a text dataset that is represented by a CSV file with columns "file", "train_label", and optionally
    "true_label". The file column should contain the path to the file that contains the actual data,
    and "train_label" refers to the label with which the data should be trained.  "true_label" refers to the actual
    label of the data point, and can differ from train_label if the dataset is poisoned.  A CSVTextDataset can support
    text data, and differs from the CSVDataset because it loads all the text data into memory and builds a vocabulary
    from it.
    """
    def __init__(self, path_to_data: str, csv_filename: str, true_label: bool = False,
                 text_field: torchtext.data.Field = None,  label_field: torchtext.data.LabelField = None,
                 max_vocab_size: int = 25000, shuffle: bool = False, random_state=None,
                 **kwargs):
        """
        Initializes the CSVTextDataset object
        :param path_to_data: root folder where all the data is located
        :param csv_filename: filename of the csv file containing the required fields to load the actual data
        :param true_label (bool): if True, then use the column "true_label" as the label associated with each
        :param text_field (torchtext.data.Field): defines how the text data will be converted to
            a Tensor.  If none, a default will be provided and tokenized with spacy
        :param label_field (torchtext.data.LabelField): defines how to process the label associated with the text
        :param max_vocab_size (int): the maximum vocabulary size that will be built
        :param shuffle: if True, the dataset is shuffled before loading into the model
        :param random_state: if specified, seeds the random sampler when shuffling the data
        :param kwargs: any additional keyword arguments, currently unused

        TODO:
         [ ] - parallelize reading in data from disk
         [ ] - revisit reading entire corpus into memory
        """

        label_column = 'train_label'
        if true_label:
            label_column = 'true_label'
        if text_field is None:
            self.text_field = torchtext.data.Field(tokenize='spacy', include_lengths=True)
            msg = "Initialized text_field to default settings with a Spacy tokenizer!"
            logger.warning(msg)
        else:
            if not isinstance(text_field, torchtext.data.Field):
                msg = "text_field must be of datatype torchtext.data.Field"
                logger.error(msg)
                raise ValueError(msg)
            self.text_field = text_field
        if label_field is None:
            self.label_field = torchtext.data.LabelField(dtype=torch.float)
            msg = ""
            logger.warning(msg)
        else:
            if not isinstance(label_field, torchtext.data.LabelField):
                msg = ""
                logger.error(msg)
                raise ValueError(msg)
            self.label_field = label_field

        self.max_vocab_size = max_vocab_size
        fields = [('text', self.text_field), ('label', self.label_field)]
        # NOTE: we read the entire dataset into memory - this may not work so well once the corpus
        # gets larger.  Revisit as necessary.
        examples = []

        # read in the specified data into the examples list
        path_to_csv = os.path.join(path_to_data, csv_filename)
        data_df = pd.read_csv(path_to_csv)
        if shuffle:
            data_df = data_df.sample(frac=1, random_state=random_state).reset_index(drop=True)

        for index, row in tqdm(data_df.iterrows(), total=len(data_df), desc="Loading Text Data..."):
            fname = row['file']
            label = row[label_column]
            with open(os.path.join(path_to_data, fname), 'r') as f:
                z = f.readlines()
                text = ' '.join(z)
            examples.append(torchtext.data.Example.fromlist([text, label], fields))

        super(CSVTextDataset, self).__init__(examples, fields, **kwargs)

    @staticmethod
    def sort_key(ex):
        return len(ex.text)



import logging
import warnings

import numpy as np
from numpy.random import RandomState

import trojai.datagen.insert_utils as insert_utils
from trojai.datagen.config import ValidInsertLocationsConfig
from .entity import Entity, GenericEntity
from .merge import Merge

logger = logging.getLogger(__name__)


"""
Module which defines several insert style merge operations.
"""


class InsertAtLocation(Merge):
    """
    Inserts a provided pattern at a specified location
    """
    def __init__(self, location: np.ndarray, protect_wrap: bool = True):
        """
        Initializes the inserter object
        :param location: The location to insert, must be of shape=(channels x 2)
        :param protect_wrap: If True, prevents insertion of objects via wrapping
        """
        self.location = location
        self.protect_wrap = protect_wrap

    def do(self, img_obj: Entity, pattern_obj: Entity, random_state_obj: RandomState) -> Entity:
        """
        Inserts a pattern into an image, using the mask of the pattern to determine which specific pixels are modifiable
        :param img_obj: The background image into which the pattern is inserted
        :param pattern_obj: The pattern to be inserted.  The mask associated with the pattern is used to determine which
                specific pixes of the pattern are inserted into the img_obj
        :param random_state_obj: ignored
        :return: The merged object
        """
        img = img_obj.get_data()
        img_mask = img_obj.get_mask()
        pattern = pattern_obj.get_data()
        pattern_mask = pattern_obj.get_mask()

        if len(img.shape) != 3:
            raise ValueError('Input image must be of dimensions rows x cols x channels')
        num_chans = img.shape[2]
        if pattern.shape[2] != num_chans:
            # force user to broadcast the pattern as necessary
            msg = 'The # of channels in the pattern does not match the # of channels in the image!'
            logger.error(msg)
            raise ValueError(msg)
        if self.location.shape[0] != num_chans:
            msg = 'location input must be of shape=(channels x 2)'
            logger.error(msg)
            raise ValueError(msg)
        if not self.protect_wrap:
            # TODO
            msg = 'Wrapping of images not yet implemented!'
            logger.error(msg)
            raise NotImplementedError(msg)

        # to allow for patterns across channels to be in different locations,
        # we do this in a for-loop
        # TODO: see if this can be vectorized
        for chan_idx in range(num_chans):
            r, c = self.location[chan_idx, :]
            chan_pattern = pattern[:, :, chan_idx].squeeze()
            p_rows, p_cols = chan_pattern.shape
            chan_location = self.location[chan_idx, :]

            logger.info("Inserting pattern into image for channel=%d at location=[%d,%d]" %
                        (chan_idx, chan_location[0], chan_location[1]))

            if self.protect_wrap:
                chan_img = img[:, :, chan_idx].squeeze()
                if not insert_utils.pattern_fit(chan_img, chan_pattern,
                                                chan_location):
                    msg = 'Pattern doesnt fit into image at specified location!'
                    logger.error(msg)
                    raise ValueError(msg)

            # take into account masks
            np.putmask(img[r:r + p_rows, c:c + p_cols, chan_idx], pattern_mask, chan_pattern)

        # TODO: is there something we need to change about the mask?
        return GenericEntity(img, img_mask)


class InsertAtRandomLocation(Merge):
    """
    Inserts a provided pattern at a random location, where valid locations are determined according to a provided
    algorithm specification
    """
    def __init__(self, method: str, algo_config: ValidInsertLocationsConfig, protect_wrap: bool = True) -> None:
        """
        Initialize the random inserter object.
        :param method: the insertion method, currently, only uniform_random_available is a valid input
        :param algo_config: The provided configuration object specifying the algorithm to use and necessary parameters
        :param protect_wrap: if True, ensures that pattern to be inserted can fit without wrapping and raises an
                             Exception otherwise
        """
        self.method = method
        self.algo_config = algo_config
        self.protect_wrap = protect_wrap

    def do(self, img_obj: Entity, pattern_obj: Entity, random_state_obj: RandomState) -> Entity:
        """
        Perform the specified merge on the input Entities and return the merged Entity
        :param img_obj: the image object into which the pattern is to be inserted
        :param pattern_obj: the pattern object to be inserted
        :param random_state_obj: used to sample from the possible valid locations, by providing a random state,
                                 we ensure reproducibility of the data
        :return: the merged Entity
        """
        pattern = pattern_obj.get_data()
        img = img_obj.get_data()
        num_chans = img.shape[2]
        if self.method == 'uniform_random_available':
            valid_location_mask = insert_utils.valid_locations(img, pattern, self.algo_config, self.protect_wrap)
            # trigger same across all channels
            if num_chans == 3:
                valid_location_mask = np.bitwise_and.reduce(valid_location_mask, axis=2)
            valid_locs = np.nonzero(valid_location_mask)
            if len(valid_locs[0]) == 0:
                # TODO: link back to this image's file pointer in error msg
                warnings.warn('Image contains no space for trigger w/out '
                              'occlusion!  Placing trigger on upper left w/ '
                              'possible partial occlusion!')
                valid_locs = np.asarray([[0, 0]] * num_chans).T
                idx_select = 0
            else:
                idx_select = random_state_obj.choice(np.arange(len(valid_locs[0])))
            logger.info("Selected random location for insertion")

            insert_locs_per_chan = np.empty((num_chans, 2), dtype=np.int16)
            for chan_idx in range(num_chans):
                insert_locs_per_chan[chan_idx, :] = [valid_locs[0][idx_select], valid_locs[1][idx_select]]
            logger.info("Inserted pattern into image")

        else:
            msg = "Insert method not yet implemented!"
            logger.error(msg)
            raise NotImplementedError(msg)

        inserter = InsertAtLocation(insert_locs_per_chan)
        inserted_img_obj = inserter.do(img_obj, pattern_obj, random_state_obj)
        return inserted_img_obj



import logging
from typing import Sequence, Dict

import skimage.transform
import numpy as np
from numpy.random import RandomState

from .entity import Entity, GenericEntity
from .transform import Transform

logger = logging.getLogger(__name__)

"""
Module defines several affine transforms using various libraries to perform  the actual transformation operation 
specified.
"""


class RotateXForm(Transform):
    """Implements a rotation of an Entity by a specified angle amount.

    """
    def __init__(self, angle: int = 90, args: tuple = (), kwargs: dict = None) -> None:
        """
        Creates a Rotator Transform object
        :param angle: The degree amount to rotate (in degrees, not radians!)
        :param args: any additional arguments to pass to skimage.transform.rotate
        :param kwargs: any keyword arguments to pass to skimage.transform.rotate
        """
        self.rotation_angle = angle
        self.args = args
        if kwargs is None:
            self.kwargs = {'preserve_range': True}
        else:
            if 'preserve_range' in kwargs and not kwargs['preserve_range']:
                msg = "preserve_range cannot be set to False!"
                logger.error(msg)
                raise ValueError(msg)
            self.kwargs = kwargs

    def do(self, input_obj: Entity, random_state_obj: RandomState) -> Entity:
        """
        Performs the rotation specified by the RotateXForm object on an input
        :param input_obj: The Entity to be rotated
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        mask = input_obj.get_mask()

        logger.info("Applying %0.02f rotation to image via skimage.transform.rotate" % (self.rotation_angle,))
        img_rotated = skimage.transform.rotate(img, self.rotation_angle, *self.args, **self.kwargs).astype(img.dtype)
        logger.info("Applying %0.02f rotation to mask via skimage.transform.rotate" % (self.rotation_angle,))
        mask_rotated = skimage.transform.rotate(mask, self.rotation_angle, *self.args, **self.kwargs)
        mask_rotated = np.logical_not(np.isclose(mask_rotated, np.zeros(mask.shape), atol=.0001))

        return GenericEntity(img_rotated, mask_rotated)


class RandomRotateXForm(Transform):
    """Implements a rotation of a random amount of degrees.

    """
    def __init__(self, angle_choices: Sequence[float] = None, angle_sampler_prob: Sequence[float] = None,
                 rotator_kwargs: Dict = None) -> None:
        """
        Creates a random rotator Transform object
        :param angle_choices: An Sequence object of floats which represent the possible angles
                              from which the sampler can choose from
        :param angle_sampler_kwargs: any keyword arguments to pass to the sampler
        """
        if angle_choices is None:
            self.angle_choices = [0, 90, 180, 270]
        else:
            self.angle_choices = angle_choices
        if rotator_kwargs is None:
            self.rotator_kwargs = {'preserve_range': True}
        else:
            self.rotator_kwargs = rotator_kwargs
        self.angle_sampler_prob = angle_sampler_prob

    def do(self, input_obj: Entity, random_state_obj: RandomState) -> Entity:
        """
        Samples from the possible angles according to the sampler specification and then applies that
        rotation to the input object
        :param input_obj: Entity to be randomly rotated
        :param random_state_obj: a random state used to maintain reproducibility through transformations
        :return: the transformed Entity
        """
        rotation_angle = random_state_obj.choice(self.angle_choices, p=self.angle_sampler_prob)

        logger.info("Sampled %0.02f rotation from RandomState" % (rotation_angle,))
        rotator = RotateXForm(rotation_angle, kwargs=self.rotator_kwargs)

        return rotator.do(input_obj, random_state_obj)



import json
import os
import re
import logging
import types

import torch
import torch.nn as nn

from .config import RunnerConfig, DefaultOptimizerConfig
from .training_statistics import TrainingRunStatistics
from .default_optimizer import DefaultOptimizer
from .optimizer_interface import OptimizerInterface

logger = logging.getLogger(__name__)


class Runner:
    """
    Fundamental unit of model generation, which trains a model as specified in a RunnerConfig object.
    """
    def __init__(self, runner_cfg: RunnerConfig,
                 persist_metadata: dict = None):
        """
        Initialize a model runner, which sets up the Optimizer, passes data to the optimizer, and collects the
        trained model and associated statistics
        :param runner_cfg: (RunnerConfig) Object that contains necessary data and objects to train a model using
            this runner.
        :param persist_metadata: (dict), if not None, the contents of this are appended to the output summary
        dictionary. This can allow for easy tracking of results if they are being collated by an additional process.
        """
        if not isinstance(runner_cfg, RunnerConfig):
            msg = "Expected a RunnerConfig object for argument 'runner_config', instead got " \
                            "type {}".format(type(runner_cfg))
            logger.error(msg)
            raise TypeError(msg)
        self.cfg = runner_cfg
        # todo: make this a type check like with runner_cfg? To reduce confusion if metadata is not a dict but code
        #   runs; make warning
        if persist_metadata is None or not isinstance(persist_metadata, dict):
            msg = "Argument 'persist_metadata' was not None nor type 'dict'. Argument will be ignored."
            logger.warning(msg)
            self.persist_info = {}
        else:
            self.persist_info = persist_metadata

    def run(self) -> None:
        """Trains a model and saves it and the associated model statistics"""
        train_data, clean_test_data, triggered_test_data = self.cfg.data.load_data()
        arch_factory_kwargs = {} if self.cfg.arch_factory_kwargs is None else self.cfg.arch_factory_kwargs

        if self.cfg.arch_factory_kwargs_generator is not None:
            arch_factory_kwargs.update(self.cfg.arch_factory_kwargs_generator(locals()))

        model = self.cfg.arch_factory.new_architecture(**arch_factory_kwargs)
        if self.cfg.parallel:
            num_available_gpus = torch.cuda.device_count()
            logger.info("Attempting to use " + str(num_available_gpus) + " GPUs for training!")
            model = nn.DataParallel(model)

        model_stats = TrainingRunStatistics()
        # TODO: this is hacked to deal w/ text data, we need to make this better
        training_cfg_list = []
        if isinstance(train_data, types.GeneratorType):
            for data, optimizer in zip(train_data, self.cfg.optimizer_generator):  # both are generators
                model, epoch_training_stats = optimizer.train(model, data, self.cfg.train_val_split)
                model_stats.add_epoch(epoch_training_stats)
                # add training configuration information to data to be saved
                training_cfg_list.append(self._get_training_cfg(optimizer))
        else:
            optimizer = next(self.cfg.optimizer_generator)
            model, epoch_training_stats = optimizer.train(model, train_data, self.cfg.train_val_split)
            model_stats.add_epoch(epoch_training_stats)
            # add training configuration information to data to be saved
            training_cfg_list.append(self._get_training_cfg(optimizer))

        # NOTE: The test function used here is one corresponding to the last optimizer used for training. An exception
        #  will be raised if no training occurred, but validation code prior to this line should prevent this from
        #  ever happening.
        test_acc = optimizer.test(model, clean_test_data, triggered_test_data)

        # Save model train/test statistics and other relevant information
        model_stats.autopopulate_final_summary_stats()
        model_stats.set_final_clean_data_test_acc(test_acc['clean_accuracy'])
        model_stats.set_final_clean_data_n_total(test_acc['clean_n_total'])
        model_stats.set_final_triggered_data_test_acc(test_acc.get('triggered_accuracy', None))
        model_stats.set_final_triggered_data_n_total(test_acc.get('triggered_n_total', None))
        self._save_model_and_stats(model, model_stats, training_cfg_list)

    @staticmethod
    def _get_training_cfg(optimizer):
        if isinstance(optimizer, DefaultOptimizerConfig):
            training_cfg = optimizer.training_cfg.get_cfg_as_dict()
        elif isinstance(optimizer, DefaultOptimizer) or isinstance(optimizer, OptimizerInterface):
            training_cfg = optimizer.get_cfg_as_dict()
        else:
            msg = "Unable to get training_cfg from optimizer(_cfg): {}, returning empty dict".format(optimizer)
            logger.warning(msg)
            training_cfg = dict()
        return training_cfg

    def _save_model_and_stats(self, model: nn.Module, stats: TrainingRunStatistics, training_cfg_list: list):
        model_path = self.cfg.model_save_dir
        if not os.path.isdir(model_path):
            try:
                os.makedirs(model_path)
            except IOError as e:
                logger.error(e)
        stats_path = self.cfg.stats_save_dir
        if not os.path.isdir(model_path):
            try:
                os.makedirs(model_path)
            except IOError as e:
                logger.error(e)

        extn = '.pt'

        if self.cfg.filename is not None:
            filename = self.cfg.filename
            if os.path.splitext(filename)[1] != extn:
                filename += extn
        else:
            if self.cfg.run_id is not None:
                filename = model.__class__.__name__ + '_id' + str(self.cfg.run_id)
            else:
                filename = model.__class__.__name__
            if self.persist_info is not None and 'name' in self.persist_info:
                filename += '_' + self.persist_info['name']
            filename += extn
        filename = self._increment_filename_if_needed(model_path, filename, extn)
        model.eval()
        model_output_fname = os.path.join(model_path, filename)
        stats_output_fname = os.path.join(stats_path, filename+'.stats.json')

        # NOTE: there are some documented issues with saving PyTorch models that were constructed
        #   with nn.DataParallel, although these seem to be resolved.  This note serves as a guide to
        #   people that if you are experiencing issues with saving models, this block of code may be
        #   the culprit.
        logger.info("Saving trained model to " + str(model_output_fname) + " in PyTorch format.")
        if self.cfg.parallel:
            model = model.module
        torch.save(model, model_output_fname)
        model_training_stats_dict = stats.get_summary()
        for i, cfg in enumerate(training_cfg_list):
            model_training_stats_dict.update({"optimizer_"+str(i): cfg})
        # add experiment configuration to the dictionary which gets printed
        model_training_stats_dict.update(self.persist_info)

        # send the statistics to the logger
        logger.info(str(model_training_stats_dict))

        # save the entire dict as a json object
        with open(stats_output_fname, 'w') as fp:
            json.dump(model_training_stats_dict, fp)

    @staticmethod
    def _increment_filename_if_needed(path: str, filename: str, extn: str) -> str:
        """
        Checks if filename already exists at path, and either adds '_1' to end of filename or increments the number
            on the end if one already is there and is separated by a non-alphanumeric value. e.g. if the file names
            'model.pt' and 'other_model_1.pt' are taken, but specified as the file names used to save the model, they
            will be saved as 'model_1.pt' and 'other_model_2.pt'.
        :param path: (str) path to directory where model is to be saved
        :param filename: (str) intended filename
        :param extn: (str) extension at which file is to be saved
        :return: (str) filename, or updated filename
        """
        if os.path.isfile(os.path.join(path, filename)):
            msg = os.path.join(path, filename) + " already exists, appending numerical id to end of file to preserve " \
                                                 "filename uniqueness!"
            logger.warning(msg)
            filename = filename[:-len(extn)]
            ending = re.search(r'[\W_]\d+$', filename)
            if ending is not None and not filename.isdigit() and ending.group()[1:] != '0':
                filename = filename[:ending.start() + 1] + str(int(ending.group()[1:]) + 1) + extn
            else:
                filename += '_1' + extn
        return filename



import logging
from abc import ABC, abstractmethod

import torch.nn

logger = logging.getLogger(__name__)


class ArchitectureFactory(ABC):
    """ Factory object that returns architectures (untrained models) for training. """
    @abstractmethod
    def new_architecture(self, **kwargs) -> torch.nn.Module:
        """
        Returns a new architecture (untrained model)
        :return: an untrained torch.nn.Module
        """
        pass

    def __eq__(self, other):
        """
        Compares two Architecture factories by comparing the string representations of the Architectures
        returned by the new_architecture() function
        :param other: the ArchitectureFactory to compare against
        :return: boolean indicating whether the architectures are the same or not
        """

        my_arch_instance = self.new_architecture()
        other_arch_instance = other.new_architecture()
        # only keep the unique elements that are not part of the nn.Module
        dir_nn_module = set(dir(torch.nn.Module))
        dir_my_arch = set(dir(my_arch_instance)) - dir_nn_module
        dir_other_arch = set(dir(other_arch_instance)) - dir_nn_module

        if len(dir_my_arch) == len(dir_other_arch):
            for item in dir_my_arch:
                if item in dir_other_arch:
                    if item[0] != '_':
                        # compare the actual objects
                        my_item = getattr(my_arch_instance, item)
                        other_item = getattr(other_arch_instance, item)
                        # NOTE: here, we check whether the arch-factory is the same based on the string representation
                        #  of a returned architecture.
                        #  this could easily be error-prone, need to revisit how to make this more robust
                        if str(my_item) != str(other_item):
                            return False
                else:
                    return False
        else:
            return False

        return True



from trojai.datagen.config import ValidInsertLocationsConfig

from typing import Sequence, Any, Tuple, Optional

import numpy as np
from scipy.ndimage import filters

import logging
logger = logging.getLogger(__name__)

# all possible directions to leave pixel along, for edge_tracing algorithm
DIRECTIONS = [(-1, -1), (-1, 1), (1, 1), (1, -1), (0, -1), (-1, 0), (0, 1), (1, 0)]


def pattern_fit(chan_img: np.ndarray, chan_pattern: np.ndarray, chan_location: Sequence[Any]) -> bool:
    """
    Returns True if the pattern at the desired location can fit into the image channel without wrap, and False otherwise

    :param chan_img: a numpy.ndarray of shape (nrows, ncols) which represents an image channel
    :param chan_pattern: a numpy.ndarray of shape (prows, pcols) which represents a channel of the pattern
    :param chan_location: a Sequence of length 2, which contains the x/y coordinate of the top left corner of the
            pattern to be inserted for this specific channel
    :return: True/False depending on whether the pattern will fit into the image
    """

    p_rows, p_cols = chan_pattern.shape
    r, c = chan_location
    i_rows, i_cols = chan_img.shape

    if (r + p_rows) > i_rows or (c + p_cols) > i_cols:
        return False

    return True


def _get_edge_length_in_direction(curr_i: int, curr_j: int, dir_i: int, dir_j: int, i_rows: int, i_cols: int,
                                  edge_pixels: set) -> int:
    """
    find the maximum length of a move in the given direction along the perimeter of the image
    :param curr_i: current row index
    :param curr_j: current col index
    :param dir_i: direction of change in row index
    :param dir_j: direction of change in col index
    :param i_rows: number of rows of containing array
    :param i_cols number of cols of containing array
    :param edge_pixels: set of remaining edge pixels to visit
    :return: the length of the edge in the given direction, 0 if none exists,
    if direction is a diagonal length will always be <= 1
    """
    length = 0
    while 0 <= curr_i + dir_i < i_rows and 0 <= curr_j + dir_j < i_cols and \
            (curr_i + dir_i, curr_j + dir_j) in edge_pixels:
        # update seen edge pixels
        edge_pixels.remove((curr_i + dir_i, curr_j + dir_j))
        length += 1
        curr_i += dir_i
        curr_j += dir_j
        # only allow length 1 diagonal moves
        if dir_i != 0 and dir_j != 0:
            break
    return length


def _get_next_edge_from_pixel(curr_i: int, curr_j: int, i_rows: int, i_cols: int,
                              edge_pixels: set) -> Optional[Tuple[int, int]]:
    """
    Obtain the next edge to trace along
    :param curr_i: current row index
    :param curr_j: current col index
    :param i_rows: number of rows of containing array
    :param i_cols: number of cols of containing array
    :param edge_pixels: set of remaining edge pixels to visit
    :return: a tuple of row distance, col distance if an undiscovered edge is found,
    otherwise None
    """
    for dir_i, dir_j in DIRECTIONS:
        length = _get_edge_length_in_direction(curr_i, curr_j, dir_i, dir_j, i_rows, i_cols, edge_pixels)
        if length != 0:
            move_i, move_j = dir_i * length, dir_j * length
            return move_i, move_j
    return None


def _get_bounding_box(coords: Sequence[int], img: np.ndarray) -> Optional[Tuple[int, int, int, int]]:
    """
    Return the smallest possible rectangle containing all non-zero pixels in img, edges inclusive
    :param coords: sequence of image subset coordinates, top, left, bottom, right
    :param img: provided image
    :return a tuple of y1 (top), x1 (left), y2 (bottom), x2 (right) of bounding box of image,
            or a 4-tuple of zeros of no non-zero pixels in image
    """
    top, left, bottom, right = coords
    img_subset = img[top:bottom, left:right]

    rows = np.logical_or.reduce(img_subset, axis=1)
    cols = np.logical_or.reduce(img_subset, axis=0)

    row_bounds = np.nonzero(rows)
    col_bounds = np.nonzero(cols)

    if row_bounds[0].size != 0 and col_bounds[0].size != 0:
        y1 = row_bounds[0][0]
        y2 = row_bounds[0][row_bounds[0].size - 1]

        x1 = col_bounds[0][0]
        x2 = col_bounds[0][col_bounds[0].size - 1]

        return top + y1, left + x1, top + y2 + 1, left + x2 + 1

    else:
        return 0, 0, 0, 0


def valid_locations(img: np.ndarray, pattern: np.ndarray, algo_config: ValidInsertLocationsConfig,
                    protect_wrap: bool = True) -> np.ndarray:
    """
    Returns a list of locations per channel which the pattern can be inserted
    into the img_channel with an overlap algorithm dictated by the appropriate
    inputs

    :param img: a numpy.ndarray which represents the image of shape:
           (nrows, ncols, nchans)
    :param pattern: the pattern to be inserted into the image of shape:
           (prows, pcols, nchans)
    :param algo_config: The provided configuration object specifying the algorithm to use and necessary parameters
    :param protect_wrap: if True, ensures that pattern to be inserted can fit without wrapping and raises an
                         Exception otherwise
    :return: A boolean mask of the same shape as the input image, with True
             indicating that that pixel is a valid location for placement of
             the specified pattern
    """
    num_chans = img.shape[2]

    # broadcast allow_overlap variable if necessary
    allow_overlap = algo_config.allow_overlap
    if not isinstance(allow_overlap, Sequence):
        allow_overlap = [allow_overlap] * num_chans
    elif len(allow_overlap) != num_chans:
        msg = "Length of provided allow_overlap sequence does not equal the number of channels in the image!"
        logger.error(msg)
        raise ValueError(msg)

    # broadcast min_val variable if necessary
    min_val = algo_config.min_val
    if not isinstance(min_val, Sequence):
        min_val = [min_val] * num_chans
    elif len(min_val) != num_chans:
        msg = "Length of provided min_val sequence does not equal the number of channels in the image!"
        logger.error(msg)
        raise ValueError(msg)

    # broadcast threshold_val variable if necessary
    threshold_val = algo_config.threshold_val
    if algo_config.algorithm == 'threshold':
        if not isinstance(threshold_val, Sequence):
            threshold_val = [threshold_val] * num_chans
        elif len(threshold_val) != num_chans:
            msg = "Length of provided threshold_val sequence does not equal the number of channels in the image!"
            logger.error(msg)
            raise ValueError(msg)

    if pattern.shape[2] != num_chans:
        # force user to broadcast the pattern as necessary
        msg = "The # of channels in the pattern does not match the # of channels in the image!"
        logger.error(msg)
        raise ValueError(msg)

    # TODO: look for vectorization opportunities
    output_mask = np.zeros(img.shape, dtype=bool)
    for chan_idx in range(num_chans):
        chan_img = img[:, :, chan_idx]
        chan_pattern = pattern[:, :, chan_idx]
        i_rows, i_cols = chan_img.shape
        p_rows, p_cols = chan_pattern.shape

        if allow_overlap[chan_idx]:
            output_mask[0:i_rows - p_rows + 1,
                        0:i_cols - p_cols + 1,
                        chan_idx] = True
        else:
            if protect_wrap:
                mask = (chan_img <= min_val[chan_idx])

                # True if image present, False if not
                img_mask = np.logical_not(mask)

                # remove boundaries from valid locations
                mask[i_rows - p_rows + 1:i_rows, :] = False
                mask[:, i_cols - p_cols + 1:i_cols] = False

                # get all edge pixels
                edge_pixels = None
                if algo_config.algorithm != 'bounding_box':
                    edge_pixel_coords = np.nonzero(
                                            np.logical_and(
                                                np.logical_xor(
                                                    filters.maximum_filter(img_mask, 3, mode='constant', cval=0.0),
                                                    filters.minimum_filter(img_mask, 3, mode='constant', cval=0.0)),
                                                img_mask))
                    edge_pixels = zip(edge_pixel_coords[0], edge_pixel_coords[1])

                if algo_config.algorithm == 'edge_tracing':
                    logger.info("Computing valid locations according to edge_tracing algorithm")
                    edge_pixel_set = set(edge_pixels)
                    # search until all edges have been visited
                    while len(edge_pixel_set) != 0:
                        start_i, start_j = edge_pixel_set.pop()

                        # invalidate relevant pixels for start square
                        top_boundary = max(0, start_i - p_rows + 1)
                        left_boundary = max(0, start_j - p_cols + 1)
                        mask[top_boundary:start_i + 1,
                             left_boundary: start_j + 1] = False

                        curr_i, curr_j = start_i, start_j
                        move = 0, 0
                        while move is not None:
                            # what edge was last traversed
                            action_i, action_j = move
                            # current location
                            curr_i += action_i
                            curr_j += action_j

                            # truncate search when near top or left boundary
                            top_index = max(0, curr_i - p_rows + 1)
                            left_index = max(0, curr_j - p_cols + 1)

                            # update invalidation based on last move, marking a row or column invalid based on the size
                            # of action_i or action_j
                            # if action_i or action_j has absolute value greater than 0, the other must be 0,
                            # i.e diagonal moves of length greater than 1 aren't updated correctly by this
                            if action_i < 0:
                                # update top border
                                mask[top_index:top_index - action_i, left_index:curr_j + 1] = False
                            elif action_i > 0:
                                # update bottom border
                                mask[curr_i - action_i + 1:curr_i + 1, left_index:curr_j + 1] = False

                            if action_j < 0:
                                # update left border
                                mask[top_index:curr_i + 1, left_index:left_index - action_j] = False
                            elif action_j > 0:
                                # update right border
                                mask[top_index:curr_i + 1, curr_j - action_j + 1:curr_j + 1] = False

                            # obtain next pixel to inspect
                            move = _get_next_edge_from_pixel(curr_i, curr_j, i_rows, i_cols, edge_pixel_set)

                elif algo_config.algorithm == 'brute_force':
                    logger.info("Computing valid locations according to brute_force algorithm")
                    for i, j in edge_pixels:
                        top_index, left_index = max(0, i - p_rows + 1), max(0, j - p_cols + 1)
                        mask[top_index:i + 1, left_index:j + 1] = False

                elif algo_config.algorithm == 'threshold':
                    logger.info("Computing valid locations according to threshold algorithm")
                    for i, j in edge_pixels:
                        mask[max(0, i - p_rows + 1):i + 1, max(0, j - p_cols + 1):j + 1] = False

                    # enumerate all possible invalid locations
                    mask_coords = np.nonzero(np.logical_not(mask))
                    possible_locations = zip(mask_coords[0], mask_coords[1])

                    # if average pixel value in location is below specified value, allow possible trigger overlap
                    for i, j in possible_locations:
                        if i <= i_rows - p_rows and j <= i_cols - p_cols and \
                                np.mean(chan_img[i:i + p_rows, j:j + p_cols]) <= threshold_val[chan_idx]:
                            mask[i][j] = True

                elif algo_config.algorithm == 'bounding_boxes':
                    logger.info("Computing valid locations according to bounding_boxes algorithm")
                    # generate top-left and bottom-right corners of all grid squares
                    top_left_coords = np.swapaxes(np.indices((algo_config.num_boxes, algo_config.num_boxes)), 0, 2) \
                                        .reshape((algo_config.num_boxes * algo_config.num_boxes, 2))
                    bottom_right_coords = top_left_coords + 1

                    # rows give y1, x1, y2, x2 of grid boxes, y2 and x2 exclusive
                    box_coords = np.concatenate((top_left_coords, bottom_right_coords), axis=1)
                    box_coords = np.multiply(box_coords, np.array([i_rows, i_cols, i_rows, i_cols]))
                    box_coords //= algo_config.num_boxes

                    # generate bounding boxes for image in each grid square
                    bounding_coords = np.apply_along_axis(_get_bounding_box, 1, box_coords, img_mask)

                    # update mask, bounds -> top, left, bottom, right
                    for bounds in bounding_coords:
                        top_index = max(0, bounds[0] - p_rows + 1)
                        left_index = max(0, bounds[1] - p_cols + 1)
                        mask[top_index:bounds[2], left_index:bounds[3]] = False

                output_mask[:, :, chan_idx] = mask

            else:
                msg = "Wrapping for trigger insertion has not been implemented yet!"
                logger.error(msg)
                raise ValueError(msg)

    return output_mask



import collections.abc
import copy
import importlib
import logging
import os
from abc import ABC, abstractmethod
from typing import Callable, Union, Sequence
from typing import Union, Sequence, Any

import cloudpickle as pickle
import numpy as np
import torch

from .architecture_factory import ArchitectureFactory
from .constants import VALID_LOSS_FUNCTIONS, VALID_DEVICES, VALID_OPTIMIZERS
from .data_manager import DataManager
from .optimizer_interface import OptimizerInterface

logger = logging.getLogger(__name__)

"""
Defines all configurations pertinent to model generation.
"""


class ConfigInterface(ABC):
    """
    Defines the interface for all configuration objects
    """

    @abstractmethod
    def __deepcopy__(self, memodict={}):
        pass


class OptimizerConfigInterface(ConfigInterface):
    @abstractmethod
    def get_device_type(self):
        pass

    def save(self, fname):
        pass

    @staticmethod
    @abstractmethod
    def load(fname):
        pass


class TrainingConfig(ConfigInterface):
    """
    Defines all required items to setup training with an optimizer
    """

    def __init__(self,
                 device: Union[str, torch.device] = 'cpu',
                 epochs: int = 10,
                 batch_size: int = 32,
                 lr: float = 1e-4,
                 optim: Union[str, OptimizerInterface] = 'adam',
                 objective: Union[str, Callable] = 'cross_entropy_loss') -> None:
        """
        Initializes a TrainingConfig object
        :param device: string or torch.device object representing the device on which computation will be performed
        :param epochs: the number of epochs to train the model
        :param batch_size: batch size used to train the model
        :param lr: the learning rate
        :param optim: either one of trojai_private.modelgen.constants.VALID_OPTIMIZERS or an optimizer
                object implementing trojai_private.modelgen.optimizer_interface.OptimizerInterface
        :param objective: either one of trojai_private.modelgen.constants.VALID_OBJECTIVES or a
                callable function that can compute a metric given y_hat and y_true
        """
        self.device = device
        self.epochs = epochs
        self.batch_size = batch_size
        self.lr = lr
        self.optim = optim
        self.objective = objective

        self.validate()

        # convert to a torch.device object
        if isinstance(self.device, str):
            self.device = torch.device(self.device)

    def validate(self) -> None:
        """
        Validate the object configuration
        :return: None
        """
        if not isinstance(self.device, torch.device) and self.device not in VALID_DEVICES:
            msg = "device must be either a torch.device object, or one of the following:" + str(VALID_DEVICES)
            logger.error(msg)
            raise ValueError(msg)
        if not isinstance(self.epochs, int) or self.epochs < 1:
            msg = "epochs must be an integer > 0"
            logger.error(msg)
            raise ValueError(msg)
        if not isinstance(self.batch_size, int) or self.batch_size < 1:
            msg = "batch_size must be an integer > 0"
            logger.error(msg)
            raise ValueError(msg)
        if not isinstance(self.lr, float):
            msg = "lr must be a float!"
            logger.error(msg)
            raise ValueError(msg)
        if not isinstance(self.optim, OptimizerInterface) and self.optim not in VALID_OPTIMIZERS:
            msg = "optim must be either a OptimizerInterface object, or one of the following:" + str(VALID_OPTIMIZERS)
            logger.error(msg)
            raise ValueError(msg)
        if not callable(self.objective) and self.objective not in VALID_LOSS_FUNCTIONS:
            msg = "objective must be a callable, or one of the following:" + str(VALID_LOSS_FUNCTIONS)
            logger.error(msg)
            raise ValueError(msg)

    def get_cfg_as_dict(self):
        """
        Returns a dictionary representation of the configuration
        :return: (dict) a dictionary
        """
        output_dict = dict(device=str(self.device.type),
                           epochs=self.epochs,
                           batch_size=self.batch_size,
                           learning_rate=self.lr,
                           optim=self.optim,
                           objective=self.objective)
        return output_dict

    def __str__(self):
        str_repr = "TrainingConfig: device[%s], num_epochs[%d], batch_size[%d], learning_rate[%.5e], optimizer[%s], " \
                   "objective[%s]" % \
                   (str(self.device.type), self.epochs, self.batch_size, self.lr,
                    str(self.optim), str(self.objective))
        return str_repr

    def __deepcopy__(self, memodict={}):
        new_device = self.device.type  # copy will keep a string version fo device, so that when
        # it gets instantiated, it will generate a device object
        # on the node
        epochs = self.epochs
        batch_size = self.batch_size
        lr = self.lr
        if isinstance(self.optim, str):
            optim = self.optim
        elif isinstance(self.optim, OptimizerInterface):
            optim = copy.deepcopy(self.optim)
        else:
            msg = "The TrainingConfig object you are trying to copy is corrupted!"
            logger.error(msg)
            raise ValueError(msg)
        if isinstance(self.objective, str):
            objective = self.objective
        elif callable(self.objective):
            objective = copy.deepcopy(self.objective)
        else:
            msg = "The TrainingConfig object you are trying to copy is corrupted!"
            logger.error(msg)
            raise ValueError(msg)
        return TrainingConfig(new_device, epochs, batch_size, lr, optim, objective)

    def __eq__(self, other):
        if self.device.type == other.device.type and self.epochs == other.epochs and \
                self.batch_size == other.batch_size and self.lr == other.lr:
            # now check the objects
            if self.optim == other.optim and self.objective == other.objective:
                return True
            else:
                return False
        else:
            return False


class ReportingConfig(ConfigInterface):
    """
    Defines all options to setup how data is reported back to the user while models are being trained
    """

    def __init__(self,
                 num_batches_per_logmsg: int = 100,
                 num_epochs_per_metric: int = 1,
                 num_batches_per_metrics: int = 50,
                 num_batches_ver_val_dataset_metrics: int = 75,
                 tensorboard_output_dir: str = '/tmp/tensorboard',
                 experiment_name: str = 'experiment'):
        """
        Initializes a ReportingConfig object.
        :param num_batches_per_logmsg: The # of batches which are computed before a log message is written.
        :param num_epochs_per_metric: The number of epochs before metrics are computed.
        :param num_batches_per_metrics: The number of batches before metrics are computed.
        :param num_batches_ver_val_dataset_metrics: The number of batches before metrics are computed for the
                validation dataset.  If None, then validation dataset metrics are not computed during training.
        :param tensorboard_output_dir: the directory to which tensorboard data should be written.
        :param experiment_name: A string identifier to associate with the configuration.
        """
        self.num_batches_per_logmsg = num_batches_per_logmsg
        self.num_epochs_per_metrics = num_epochs_per_metric
        self.num_batches_per_metrics = num_batches_per_metrics
        self.num_batches_per_val_dataset_metrics = num_batches_ver_val_dataset_metrics
        self.tensorboard_output_dir = tensorboard_output_dir
        self.experiment_name = experiment_name

        self.validate()

    def validate(self):
        if not isinstance(self.num_batches_per_logmsg, int) or self.num_batches_per_logmsg < 0:
            msg = "num_batches_per_logmsg must be an integer > 0"
            logger.error(msg)
            raise ValueError(msg)
        if not isinstance(self.num_epochs_per_metrics, int) or self.num_epochs_per_metrics < 0:
            msg = "num_epochs_per_metrics must be an integer > 0"
            logger.error(msg)
            raise ValueError(msg)
        if not isinstance(self.num_batches_per_metrics, int) or self.num_batches_per_metrics < 0:
            msg = "num_batches_per_metrics must be an integer > 0"
            logger.error(msg)
            raise ValueError(msg)
        if self.num_batches_per_val_dataset_metrics is not None:
            if not isinstance(self.num_batches_per_val_dataset_metrics,
                              int) or self.num_batches_per_val_dataset_metrics < 0:
                msg = "num_batches_per_val_dataset_metrics must be an integer > 0 or None"
                logger.error(msg)
                raise ValueError(msg)

    def __str__(self):
        str_repr = "ReportingConfig: num_batches/log_msg[%d], num_epochs/metric[%d], num_batches/metric[%d], " \
                   "num_batches/val_dataset_metric[%d] tensorboard_dir[%s] experiment_name=[%s]" % \
                   (self.num_batches_per_logmsg, self.num_epochs_per_metrics, self.num_batches_per_metrics,
                    self.num_batches_per_val_dataset_metrics, self.tensorboard_output_dir, self.experiment_name)
        return str_repr

    def __copy__(self):
        return ReportingConfig(self.num_batches_per_logmsg, self.num_epochs_per_metrics,
                               self.num_batches_per_metrics, self.num_batches_per_val_dataset_metrics,
                               self.tensorboard_output_dir, self.experiment_name)

    def __deepcopy__(self, memodict={}):
        return self.__copy__()

    def __eq__(self, other):
        if self.num_batches_per_logmsg == other.num_batches_per_logmsg and \
                self.num_epochs_per_metrics == other.num_epochs_per_metrics and \
                self.num_batches_per_metrics == other.num_batches_per_metrics and \
                self.num_batches_per_val_dataset_metrics == other.num_batches_per_val_dataset_metrics and \
                self.tensorboard_output_dir == other.tensorboard_output_dir and \
                self.experiment_name == other.experiment_name:
            return True
        else:
            return False


class LSTMOptimizerConfig(OptimizerConfigInterface):
    """
    Defines the configuration needed to setup the LSTMOptimizer
    """

    def __init__(self, training_cfg: TrainingConfig = None, reporting_cfg: ReportingConfig = None):
        """
        Initializes a LSTM Optimizer
        :param training_cfg: a TrainingConfig object, if None, a default TrainingConfig object will be constructed
        :param reporting_cfg: a ReportingConfig object, if None, a default ReportingConfig object will be constructed
        """
        if training_cfg is None:
            logger.info("Using default training configuration to setup Optimizer!")
            self.training_cfg = TrainingConfig()
        elif not isinstance(training_cfg, TrainingConfig):
            msg = "training_cfg must be of type TrainingConfig"
            logger.error(msg)
            raise TypeError(msg)
        else:
            self.training_cfg = training_cfg

        if reporting_cfg is None:
            logger.info("Using default reporting configuration to setup Optimizer!")
            self.reporting_cfg = ReportingConfig()
        elif not isinstance(reporting_cfg, ReportingConfig):
            msg = "reporting_cfg must be of type ReportingConfig"
            logger.error(msg)
            raise TypeError(msg)
        else:
            self.reporting_cfg = reporting_cfg

    def __deepcopy__(self, memodict={}):
        training_cfg_copy = copy.deepcopy(self.training_cfg)
        reporting_cfg_copy = copy.deepcopy(self.reporting_cfg)
        return LSTMOptimizerConfig(training_cfg_copy, reporting_cfg_copy)

    def __eq__(self, other):
        if self.training_cfg == other.training_cfg and self.reporting_cfg == other.reporting_cfg:
            return True
        else:
            return False

    def save(self, fname):
        """
        Saves the optimizer configuration to a file
        :param fname: the filename to save the config to
        :return: None
        """
        with open(fname, 'wb') as f:
            pickle.dump(self, f)

    @staticmethod
    def load(fname):
        """
        Loads a configuration from disk
        :param fname: the filename where the config is stored
        :return: the loaded configuration
        """
        with open(fname, 'rb') as f:
            loaded_optimzier_cfg = pickle.load(f)
        return loaded_optimzier_cfg

    def get_device_type(self):
        """
        Returns the device associated w/ this optimizer configuration.  Needed to save/load for UGE.
        :return (str): the device type represented as a string
        """
        return str(self.training_cfg.device)


class DefaultOptimizerConfig(OptimizerConfigInterface):
    """
    Defines the configuration needed to setup the DefaultOptimizer
    """

    def __init__(self, training_cfg: TrainingConfig = None, reporting_cfg: ReportingConfig = None):
        """
        Initializes a Default Optimizer
        :param training_cfg: a TrainingConfig object, if None, a default TrainingConfig object will be constructed
        :param reporting_cfg: a ReportingConfig object, if None, a default ReportingConfig object will be constructed
        """
        if training_cfg is None:
            logger.info("Using default training configuration to setup Optimizer!")
            self.training_cfg = TrainingConfig()
        elif not isinstance(training_cfg, TrainingConfig):
            msg = "training_cfg must be of type TrainingConfig"
            logger.error(msg)
            raise TypeError(msg)
        else:
            self.training_cfg = training_cfg

        if reporting_cfg is None:
            logger.info("Using default reporting configuration to setup Optimizer!")
            self.reporting_cfg = ReportingConfig()
        elif not isinstance(reporting_cfg, ReportingConfig):
            msg = "reporting_cfg must be of type ReportingConfig"
            logger.error(msg)
            raise TypeError(msg)
        else:
            self.reporting_cfg = reporting_cfg

    def __deepcopy__(self, memodict={}):
        training_cfg_copy = copy.deepcopy(self.training_cfg)
        reporting_cfg_copy = copy.deepcopy(self.reporting_cfg)
        return DefaultOptimizerConfig(training_cfg_copy, reporting_cfg_copy)

    def __eq__(self, other):
        if self.training_cfg == other.training_cfg and self.reporting_cfg == other.reporting_cfg:
            return True
        else:
            return False

    def get_device_type(self):
        """
        Returns the device associated w/ this optimizer configuration.  Needed to save/load for UGE.
        :return (str): the device type represented as a string
        """
        return str(self.training_cfg.device)

    def save(self, fname):
        """
        Saves the optimizer configuration to a file
        :param fname: the filename to save the config to
        :return: None
        """
        with open(fname, 'wb') as f:
            pickle.dump(self, f)

    @staticmethod
    def load(fname):
        """
        Loads a configuration from disk
        :param fname: the filename where the config is stored
        :return: the loaded configuration
        """
        with open(fname, 'rb') as f:
            loaded_optimzier_cfg = pickle.load(f)
        return loaded_optimzier_cfg


class ModelGeneratorConfig(ConfigInterface):
    """Object used to configure the model generator"""

    def __init__(self, arch_factory: ArchitectureFactory, data: DataManager,
                 model_save_dir: str, stats_save_dir: str, num_models: int,
                 arch_factory_kwargs: dict = None, arch_factory_kwargs_generator: Callable = None,
                 optimizer: Union[Union[OptimizerInterface, DefaultOptimizerConfig],
                                  Sequence[Union[OptimizerInterface, DefaultOptimizerConfig]]] = None,
                 parallel=False,
                 train_val_split=0.,
                 experiment_cfg: dict = None,
                 run_ids: Union[Any, Sequence[Any]] = None,
                 filenames: Union[str, Sequence[str]] = None):
        """
        Initializes the ModelGeneratorConfig object which provides needed information for generating models for a given
        experiment.

        :param arch_factory: ArchitectureFactory object that provides instantiated
            architectures (untrained models) to be trained on the data.
        :param data: TrojaiDataManager object containing the experiment path and files.
        :param model_save_dir: path to directory where the models should be saved
        :param stats_save_dir: path to directory where the model training stats should be saved
        :param num_models: number of models to train with this configuration
        :param arch_factory_kwargs: (dict) a dictionary which contains keywords and associated values
            that are needed to instantiate a trainable module from the factory
        :param arch_factory_kwargs_generator: (callable) a callable, or None, which takes a dictionary of all
            variables defined in the Runner's namespace, and then creates a new dictionary that contains the keyword
            arguments to instantiate an architecture from the architecture factory
        :param optimizer: a OptimizerInterface object, or a DefaultOptimizer configuration, or possibly mixed sequence
            of both.  If a sequence of optimizers is passed, then the length of that sequence must match the number
            of sequential datasets that are to be used for training the model.
        :param parallel: (bool) - if True, attempts to use multiple GPU's
        :param train_val_split: (float) if > 0, then splits the training dataset and uses it as validation.  If 0
            the training dataset is not split and validation is not computed
        :param experiment_cfg: dictionary containing information regarding the experiment which is being run by the
            ModelGenerator.  This information is also saved in the output summary JSON file that is associated with
            every model that is generated.
        :param run_ids: Identifiers for models. If a sequence, len(run_ids) must be equal to num_models
        :param filenames: An optional list of file names to save each model by each
            file name, or a single filename to have models be saved with the same file name with '_#' added to
            the end, e.g. 'filename.pt', 'filename_1.pt', 'filename_2.pt', ...
            If this argument is not provided, then models generated will be saved with filenames indicated by the
            experiment name in the experiment_cfg dictionary
        """
        self.arch_factory = arch_factory
        self.arch_factory_kwargs = arch_factory_kwargs
        self.arch_factory_kwargs_generator = arch_factory_kwargs_generator
        self.data = data
        self.model_save_dir = model_save_dir
        self.stats_save_dir = stats_save_dir
        self.num_models = num_models

        self.optimizer = optimizer
        self.parallel = parallel
        self.train_val_split = train_val_split
        self.experiment_cfg = dict() if experiment_cfg is None else experiment_cfg

        self.run_ids = run_ids  # it might be useful to allow something like a generator for this argument
        self.filenames = filenames  # it might be useful to allow something like a generator for this argument

        self.validate()

    def __deepcopy__(self, memodict={}):
        arch_factory_copy = copy.deepcopy(
            self.arch_factory)  # I think this is OK b/c the ArchFactory is a class definition
        data_copy = copy.deepcopy(self.data)  # the default should work properly here b/c all properties are primitives
        optimizer_copy = copy.deepcopy(self.optimizer)
        return ModelGeneratorConfig(arch_factory_copy, data_copy,
                                    self.model_save_dir, self.stats_save_dir, self.num_models,
                                    self.arch_factory_kwargs, self.arch_factory_kwargs_generator,
                                    optimizer_copy, self.parallel, self.train_val_split, self.experiment_cfg,
                                    self.run_ids, self.filenames)

    def __eq__(self, other):
        if self.arch_factory == other.arch_factory and self.data == other.data and self.optimizer == other.optimizer \
                and self.parallel == other.parallel and self.train_val_split == other.train_val_split \
                and self.model_save_dir == other.model_save_dir and self.stats_save_dir == other.stats_save_dir \
                and self.arch_factory_kwargs == other.arch_factory_kwargs \
                and self.arch_factory_kwargs_generator == other.arch_factory_kwargs_generator \
                and self.experiment_cfg == other.experiment_cfg and self.run_ids == other.run_ids \
                and self.filenames == other.filenames:
            return True
        else:
            return False

    def validate(self) -> None:
        """
        Validate the input arguments to construct the object
        :return: None
        """
        if not (isinstance(self.arch_factory, ArchitectureFactory)):
            msg = "Expected an ArchitectureFactory object for argument 'architecture_factory', " \
                  "instead got type: {}".format(type(self.arch_factory))
            logger.error(msg)
            raise TypeError(msg)
        if self.arch_factory_kwargs is not None and not isinstance(self.arch_factory_kwargs, dict):
            msg = "Expected dictionary for arch_factory_kwargs"
            logger.error(msg)
            raise TypeError(msg)
        if self.arch_factory_kwargs_generator is not None and not callable(self.arch_factory_kwargs_generator):
            msg = "arch_factory_kwargs_generator must be a Callable!"
            logger.error(msg)
            raise TypeError(msg)
        if not (isinstance(self.data, DataManager)):
            msg = "Expected an TrojaiDataManager object for argument 'data', " \
                  "instead got type: {}".format(type(self.data))
            logger.error(msg)
            raise TypeError(msg)
        if not type(self.model_save_dir) == str:
            msg = "Expected type 'string' for argument 'model_save_dir, instead got type: " \
                  "{}".format(type(self.model_save_dir))
            logger.error(msg)
            raise TypeError(msg)
        if not os.path.isdir(self.model_save_dir):
            try:
                os.makedirs(self.model_save_dir)
            except IOError as e:
                msg = "'model_save_dir' was not found and could not be created" \
                      "...\n{}".format(e.__traceback__)
                logger.error(msg)
                raise IOError(msg)
        if not type(self.num_models) == int:
            msg = "Expected type 'int' for argument 'num_models, instead got type: " \
                  "{}".format(type(self.num_models))
            logger.error(msg)
            raise TypeError(msg)
        if self.filenames is not None:
            if isinstance(self.filenames, Sequence):
                for filename in self.filenames:
                    if not type(filename) == str:
                        msg = "Encountered non-string in argument 'filenames': {}".format(filename)
                        logger.error(msg)
                        raise TypeError(msg)
            else:
                if not isinstance(self.filenames, str):
                    msg = "Filename provided as prefix must be of type string!"
                    logger.error(msg)
                    raise TypeError(msg)
        if self.run_ids is not None and len(self.run_ids) != self.num_models:
            msg = "Argument 'run_ids' was provided, but len(run_ids) != num_models"
            logger.error(msg)
            raise RuntimeError(msg)
        if self.filenames is not None and len(self.filenames) != self.num_models:
            msg = "Argument 'filenames' was provided, but len(filenames) != num_models"
            logger.error(msg)
            raise RuntimeError(msg)
        if self.run_ids is not None and self.filenames is not None:
            msg = "Argument 'filenames' was provided with argument 'run_ids', 'run_ids' will be ignored..."
            logger.warning(msg)

        RunnerConfig.validate_optimizer(self.optimizer, self.data)

        if not isinstance(self.parallel, bool):
            msg = "parallel argument must be a boolean!"
            logger.error(msg)
            raise ValueError(msg)
        if not isinstance(self.train_val_split, float):
            msg = "train_val_split must a float between 0 and 1!"
            logger.error(msg)
            raise ValueError(msg)
        else:
            if self.train_val_split < 0 or self.train_val_split > 1:
                msg = "train_val_split must be between 0 and 1, inclusive"
                logger.error(msg)
                raise ValueError(msg)

    def __getstate__(self):
        """
        Function which dictates which objects will be saved when pickling the ModelGeneratorConfig object.  This is
        only useful for the UGEModelGenerator, which needs to save the data before parallelizing a job.
        :return: a dictionary of the state of the ModelGeneratorConfig object.
        """
        return {'arch_factory': self.arch_factory,
                'data': self.data,
                'model_save_dir': self.model_save_dir,
                'stats_save_dir': self.stats_save_dir,
                'num_models': self.num_models,
                'arch_factory_kwargs': self.arch_factory_kwargs,
                'arch_factory_kwargs_generator': self.arch_factory_kwargs_generator,
                'parallel': self.parallel,
                'train_val_split': self.train_val_split,
                'experiment_cfg': self.experiment_cfg,
                'run_ids': self.run_ids,
                'filenames': self.filenames
                }

    def save(self, fname: str):
        """
        Saves the ModelGeneratorConfig object in two different parts.  Every object within the config, except for the
        optimizer is saved in the .klass.save file, and the optimizer is saved separately.
        :param fname - the filename to save the configuration to
        :return: None
        """
        # we save optimizer and the remainder of the components separately
        optimizer_klass_save_fname = fname + '.optimizer.klass.save'
        optimizer_save_fname = fname + '.optimizer.save'
        remainder_data_save_fname = fname + '.arch_data.save'

        with open(remainder_data_save_fname, 'wb') as f:
            pickle.dump(self, f)
        # save the optimizer class name, so we can load it properly
        optimizer_klass_name = '.'.join([self.optimizer.__module__, self.optimizer.__class__.__name__])
        with open(optimizer_klass_save_fname, 'w') as f:
            f.write(optimizer_klass_name)
        self.optimizer.save(optimizer_save_fname)

    @staticmethod
    def load(fname: str):
        """
        Loads a saved modelgen_cfg object from data that was saved using the .save() function.
        :param fname: the filename where the modelgen_cfg object is saved
        :return: a ModelGeneratorConfig object
        """
        optimizer_klass_save_fname = fname + '.optimizer.klass.save'
        optimizer_save_fname = fname + '.optimizer.save'
        remainder_data_save_fname = fname + '.arch_data.save'

        with open(remainder_data_save_fname, 'rb') as f:
            modelgen_cfg = pickle.load(f)

        # load the class name of the optimizer that was used
        with open(optimizer_klass_save_fname, 'r') as f:
            optimizer_module_and_klass_name = f.readline()
        # load the module
        ss = optimizer_module_and_klass_name.split('.')
        optimizer_module_name = '.'.join(ss[0:-1])
        optimizer_klass_name = ss[-1]
        optimizer_module = importlib.import_module(optimizer_module_name)
        optimizer_klass_def = getattr(optimizer_module, optimizer_klass_name)
        optimizer_load = optimizer_klass_def.load(optimizer_save_fname)

        # reconstruct the ModelGeneratorConfig object
        modelgen_cfg.optimizer = optimizer_load
        modelgen_cfg.validate()
        return modelgen_cfg


class RunnerConfig(ConfigInterface):
    """
    Container for all parameters needed to use the Runner to train a model.
    """

    def __init__(self, arch_factory: ArchitectureFactory, data: DataManager,
                 arch_factory_kwargs: dict = None, arch_factory_kwargs_generator: Callable = None,
                 optimizer: Union[OptimizerInterface, DefaultOptimizerConfig,
                                  Sequence[Union[OptimizerInterface, DefaultOptimizerConfig]]] = None,
                 parallel: bool = False, train_val_split=0.,
                 model_save_dir: str = "/tmp/models", stats_save_dir: str = "/tmp/model_stats",
                 run_id: Any = None, filename: str = None):
        """
        Initialize a RunnerConfig object
        :param arch_factory: (Architecture Factory) a trainable Pytorch module generator.
        :param data: (TrojaiDataManager) a TrojaiDataManager object containing the paths to the data being trained and
            tested on, as well as functions dictating how the data should be transformed for training and testing.
        :param arch_factory_kwargs: (dict) a dictionary which contains keywords and associated values
            that are needed to instantiate a trainable module from the factory
        :param arch_factory_kwargs_generator: (callable) a callable, or None, which takes a dictionary of all
            variables defined in the Runner's namespace, and then creates a new dictionary taht contains the keyword
            arguments to instantiate an architecture from the architecture factory
        :param optimizer: a OptimizerInterface object, or a DefaultOptimizer configuration, or possibly mixed sequence
            of both
        :param parallel: (bool) if True, spreads GPU tasking over all available GPUs
        :param train_val_split: (float) if > 0, then splits the training dataset and uses it as validation.  If 0
            the training dataset is not split and validation is not computed
        :param model_save_dir: (str) path to where the models should be saved.
        :param stats_save_dir: (str) path to where the model training statistics should be saved.
        :param run_id: An ending to the save file name. Can be anything, but will be converted to string format.
            Ignored if a filename is provided.
        :param filename: (str) File name for the saved model. If not specified, default to the name of the architecture
            provided. Should end in .pt for consistency.
        """
        self.arch_factory = arch_factory
        self.data = data
        self.arch_factory_kwargs = arch_factory_kwargs
        self.arch_factory_kwargs_generator = arch_factory_kwargs_generator
        self.optimizer = optimizer
        self.parallel = parallel
        self.train_val_split = train_val_split
        self.model_save_dir = model_save_dir
        self.stats_save_dir = stats_save_dir
        self.run_id = run_id
        self.filename = filename

        self.validate()

        # create new attribute instead of overwriting self.optimizer so that self.__deepcopy__ still works.
        self.optimizer_generator = self.setup_optimizer_generator(self.optimizer, self.data)

    def __deepcopy__(self, memodict={}):
        arch_copy = copy.deepcopy(self.arch_factory)
        data_copy = copy.deepcopy(self.data)
        optim_copy = copy.deepcopy(self.optimizer)
        return RunnerConfig(arch_copy, data_copy, self.arch_factory_kwargs, self.arch_factory_kwargs_generator,
                            optim_copy, self.parallel, self.train_val_split,
                            self.model_save_dir, self.stats_save_dir,
                            self.run_id, self.filename)

    @staticmethod
    def setup_optimizer_generator(optimizer, data):
        """
        Converts an optimizer specification to a generator, to be compatible with sequential training.
        :param optimizer: the optimizer to configure into a generator
        :param num_datasets: the number of datasets for which optimizers need to be created
        :return: A generator that returns optimizers for every dataset to be trained
        """
        from .default_optimizer import DefaultOptimizer
        if optimizer is None or isinstance(optimizer, DefaultOptimizerConfig):
            return (DefaultOptimizer(optimizer) for _ in range(len(data.train_file)))
        elif isinstance(optimizer, OptimizerInterface):
            return (optimizer.__deepcopy__({}) for _ in range(len(data.train_file)))
        else:
            msg = "Multiple optimizers specified, only final will be used for test calculations"
            logger.warning(msg)
            return (opt if isinstance(opt, OptimizerInterface) else DefaultOptimizer(opt) for opt in optimizer)

    @staticmethod
    def validate_optimizer(optimizer, data):
        """
        Validates an optimzer configuration
        :param optimizer: the optimizer/optimizer configuration to be validated
        :param data: the data to be optimized
        :return:
        """
        if not (optimizer is None
                or isinstance(optimizer, OptimizerInterface)
                or isinstance(optimizer, DefaultOptimizerConfig)):
            if not (hasattr(type(optimizer), '__iter__') and hasattr(type(optimizer), '__len__') and
                    type(optimizer) != str):
                msg = "Expected OptimizerInterface, DefaultOptimizerConfig, or sequence of them for argument" \
                      "'optimizer', instead got {}".format(optimizer)
                logger.error(msg)
                raise TypeError(msg)
            else:
                for item in optimizer:
                    if not (isinstance(item, OptimizerInterface) or isinstance(item, DefaultOptimizerConfig)):
                        msg = "Expected OptimizerInterface or DefaultOptimizerConfig objects in sequence for argument" \
                              "'optimizer', discovered {} in sequence".format(item)
                        logger.error(msg)
                        raise TypeError(msg)
                if len(optimizer) != len(data.train_file):
                    msg = "If specifying multiple optimizers, the number of optimizers given must be the same as the " \
                          "number of training files in the DataManager."
                    logger.error(msg)
                    raise TypeError(msg)

    def validate(self) -> None:
        """
        Validate the RunnerConfig object
        :return: None
        """
        if not isinstance(self.arch_factory, ArchitectureFactory):
            msg = "Expected ArchitectureFactory for argument 'architecture', instead got type: {}".format(type(
                self.arch_factory))
            logger.error(msg)
            raise TypeError(msg)
        if self.arch_factory_kwargs is not None and not isinstance(self.arch_factory_kwargs, dict):
            msg = "arch_factory_kwargs must be a dictionary!"
            logger.error(msg)
            raise TypeError(msg)
        if self.arch_factory_kwargs_generator is not None and not callable(self.arch_factory_kwargs_generator):
            msg = "Expected a function for argument 'arch_factory_kwargs_generator', " \
                  "instead got type: {}".format(type(self.arch_factory_kwargs_generator))
            logger.error(msg)
            raise TypeError(msg)
        if not isinstance(self.data, DataManager):
            msg = "Expected a TrojaiDataManager object for argument 'data', " \
                  "instead got type: {}".format(type(self.data))
            logger.error(msg)
            raise TypeError(msg)

        self.validate_optimizer(self.optimizer, self.data)
        if not isinstance(self.parallel, bool):
            msg = "parallel argument must be a boolean!"
            logger.error(msg)
            raise ValueError(msg)
        if not isinstance(self.train_val_split, float):
            msg = "train_val_split must a float between 0 and 1!"
            logger.error(msg)
            raise ValueError(msg)
        else:
            if self.train_val_split < 0 or self.train_val_split > 1:
                msg = "train_val_split must be between 0 and 1, inclusive"
                logger.error(msg)
                raise ValueError(msg)

        if not type(self.model_save_dir) == str:
            msg = "Expected type 'string' for argument 'model_save_dir, instead got type: " \
                  "{}".format(type(self.model_save_dir))
            logger.error(msg)
            raise TypeError(msg)
        if not os.path.isdir(self.model_save_dir):
            try:
                os.makedirs(self.model_save_dir)
            except OSError as e:  # not sure this error is possible as written
                msg = "'model_save_dir' was not found and could not be created" \
                      "...\n{}".format(e.__traceback__)
                logger.error(msg)
                raise OSError(msg)
        if not os.path.isdir(self.stats_save_dir):
            try:
                os.makedirs(self.stats_save_dir)
            except OSError as e:  # not sure this error is possible as written
                msg = "'stats_save_dir' was not found and could not be created" \
                      "...\n{}".format(e.__traceback__)
                logger.error(msg)
                raise OSError(msg)
        if self.filename is not None and not type(self.filename) == str:
            msg = "Expected a string for argument 'filename', instead got " \
                  "type {}".format(type(self.filename))
            logger.error(msg)
            raise TypeError(msg)


def modelgen_cfg_to_runner_cfg(modelgen_cfg: ModelGeneratorConfig,
                               run_id=None,
                               filename=None) -> RunnerConfig:
    """
    Convenience function which creates a RunnerConfig object, from a ModelGeneratorConfig object.
    :param modelgen_cfg: the ModelGeneratorConfig to convert
    :param run_id: run_id to be associated with the RunnerConfig
    :param filename: filename to be associated with the RunnerConfig
    :return: the created RunnerConfig object
    """
    return RunnerConfig(modelgen_cfg.arch_factory, modelgen_cfg.data, modelgen_cfg.arch_factory_kwargs,
                        modelgen_cfg.arch_factory_kwargs_generator,
                        modelgen_cfg.optimizer, modelgen_cfg.parallel, modelgen_cfg.train_val_split,
                        modelgen_cfg.model_save_dir, modelgen_cfg.stats_save_dir,
                        run_id=run_id, filename=filename)


class UGEQueueConfig:
    """
    Defines the configuration for a Queue w.r.t. UGE in TrojAI
    """
    def __init__(self, queue_name: str, gpu_enabled: bool, sync_mode: bool = False):
        self.queue_name = queue_name
        self.gpu_enabled = gpu_enabled
        self.sync_mode = sync_mode

    def validate(self) -> None:
        """
        Validate the UGEQueueConfig object
        """
        if not isinstance(self.queue_name, str):
            msg = "queue_name must be a string!"
            logger.error(msg)
            raise TypeError(msg)
        if not isinstance(self.gpu_enabled, bool):
            msg = "gpu_enabled argument must be a boolean!"
            logger.error(msg)
            raise TypeError(msg)
        if not isinstance(self.sync_mode, bool):
            msg = "sync_mode argument must be a boolean!"
            logger.error(msg)
            raise TypeError(msg)
        if self.sync_mode:
            msg = "sync_mode=True currently unsupported!"
            logger.error(msg)
            raise TypeError(msg)


class UGEConfig:
    """
    Defines a configuration for the UGE
    """
    def __init__(self, queues: Union[UGEQueueConfig, Sequence[UGEQueueConfig]],
                 queue_distribution: Sequence[float] = None,
                 multi_model_same_gpu: bool = False):
        """
        :param queues: a list of Queue object configurations
        :param queue_distribution: the desired way to distribute the workload across the queues, if None,
                then the workload is distributed evenly across the queues, otherwise
        :param multi_model_same_gpu: if True, then if multiple models are desired for a given ModelGeneratorConfig,
                those will all be trained on the same queue.  Otherwise, they will be distributed as much as possible
                (which is likely to complete the job faster!)
        """
        self.queues = queues
        self.queue_distribution = queue_distribution
        self.multi_model_same_gpu = multi_model_same_gpu
        self.validate()

    def validate(self):
        """
        Validate the UGEConfig object
        """
        if isinstance(self.queues, UGEQueueConfig):
            self.queues = [self.queues]
        elif isinstance(self.queues, collections.abc.Sequence):
            for q in self.queues:
                if not isinstance(q, UGEQueueConfig):
                    msg = "queues must be a Sequence of UGEQueueConfig objects!"
                    logger.error(msg)
                    raise TypeError(msg)
        else:
            msg = "queues input must be either a UGEQueueConfig object, or a Sequence of UGEQueueConfig objects!"
            logger.error(msg)
            raise TypeError(msg)

        if self.queue_distribution is not None:
            if not isinstance(self.queue_distribution, collections.abc.Sequence):
                msg = "queue_distribution argument must be either None (implying uniform distribution among all " \
                      "queues, or a Sequence of floats summing to one"
                logger.error(msg)
                raise TypeError(msg)
            else:
                try:
                    if len(self.queue_distribution) != len(self.queues):
                        msg = "if a queue_distribution is provided, it must be equal to the number of queues provided!"
                        logger.error(msg)
                        raise TypeError(msg)
                    sum_val = np.sum(self.queue_distribution)
                    if not np.isclose(sum_val, 1):
                        msg = "queue_distribution must be a Sequence of floats summing to 1"
                        logger.error(msg)
                        raise ValueError(msg)
                    for d in self.queue_distribution:
                        if d < 0 or d > 1:
                            msg = "queue_distribution values must be between 0 and 1"
                            logger.error(msg)
                            raise TypeError(msg)
                except TypeError as e:
                    logger.exception(e)
                    raise TypeError(e)

        if not isinstance(self.multi_model_same_gpu, bool):
            msg = "multi_model_same_gpu input must be a boolean!"
            logger.error(msg)
            raise TypeError(msg)


import cv2
import numpy as np
from typing import Tuple, Optional

import logging
logger = logging.getLogger(__name__)

"""
Contains general utilities for dealing with channel formats
"""


def gray_to_rgb(img: np.ndarray) -> np.ndarray:
    """
    Convert given grayscale image to RGB
    :param img: 1-channel grayscale image
    :return: image converted to RGB
    """
    if not (len(img.shape) == 2 or (len(img.shape) == 3 and img.shape[2] == 1)):
        raise TypeError("input image must be in either shape (rows, cols) or (rows, cols, 1)!")

    return cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)


def rgba_to_rgb(img: np.ndarray) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Split given 4-channel RGBA array into a 3-channel RGB array and a 1-channel alpha array
    :param img: given image to split, must be 3-channel or 4-channel
    :return: the first three channels of data as a 3-channel RGB image and the fourth channel of img as either a
    1-channel alpha array, or None if img has only 3 channels
    """
    if len(img.shape) != 3:
        raise TypeError('Input image must be in shape (rows, cols, channels)!')

    if img.shape[2] == 3:
        return img, None
    elif img.shape[2] == 4:
        r_ch, g_ch, b_ch, alpha_ch = cv2.split(img)
        return cv2.merge((r_ch, g_ch, b_ch)), alpha_ch
    else:
        raise TypeError("rgba_to_rgb expects a 3-channel or 4-channel input image,"
                        "%d-channel input image was detected!" % img.shape[2])


def rgb_to_rgba(img, alpha_ch: Optional[np.ndarray] = None) -> np.ndarray:
    """
    Converts given image to RGBA, with optionally provided alpha_ch as its alpha channel
    :param img: 3-channel RGB image or 4-channel RGBA image
    :param alpha_ch: 1-channel array to be used as alpha value (optional),
    if img is RGBA this value is ignored
    :return: if img is 4-channel it is returned unmodified, if img is 3-channel this will return a new RGBA image with
    img as its RGB channels and either alpha_ch as its alpha channel if provided or a fully opaque alpha channel
    (max value for its datatype)
    """
    if len(img.shape) != 3:
        raise TypeError('Input image must be in shape (rows, cols, channels)!')

    if img.shape[2] == 3:
        if alpha_ch is None:
            return cv2.cvtColor(img, cv2.COLOR_RGB2RGBA)
        else:
            if len(alpha_ch.shape) == 2:
                return cv2.merge((img, alpha_ch))
            else:
                raise TypeError("input alpha channel to be must in shape (rows, cols)!")
    elif img.shape[2] == 4:
        return img
    else:
        raise TypeError("rgb_to_rgba expects a 3-channel or 4-channel input image, "
                        "%d-channel input image was detected!" % img.shape[2])


def normalization_to_rgb(img: np.ndarray, normalize: bool, name: str) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Guard for input to RGB only xforms
    :param img: input image with variable number of channels
    :param normalize: whether to attempt to convert img from original channel format to 3-channel RGB
    :param name: name of calling xform
    :return: a 3-channel RGB array converted from img,
    additional conversions can be added below, currently only RGBA to RGB is implemented
    """
    if len(img.shape) != 3:
        raise TypeError('Input image must be in shape (rows, cols, channels)!')

    original_n_chan = img.shape[2]
    if original_n_chan == 3:
        return img, None

    if normalize:
        if original_n_chan == 4:
            return rgba_to_rgb(img)
        else:
            raise TypeError("No conversion for %d-channel to 3-channel images is implemented!" % original_n_chan)
    else:
        raise TypeError("%s is an RGB-only transform, %d-channel input was detected!" % (name, original_n_chan,))


def normalization_from_rgb(rgb_img: np.ndarray, alpha_ch: Optional[np.ndarray], normalize: bool, original_n_chan: int,
                           name: str) -> np.ndarray:
    """
    Guard for output from rgb-only xforms
    :param rgb_img: 3-channel RGB image result from calling xform
    :param alpha_ch: alpha channel extracted at beginning of calling xform or None
    :param normalize: whether to convert rgb_img back to its original channel format
    :param original_n_chan: number of channels in its original channel format
    :param name: name of calling xform
    :return: if normalize is True the image corresponding to rgb_img converted to its original channel format,
    otherwise rgb_img unmodified,
    additional conversions can be added below, currently only RGB to RGBA is implemented
    """
    if len(rgb_img.shape) != 3:
        raise TypeError('Input image must be in shape (rows, cols, channels)!')

    if rgb_img.shape[2] != 3:
        raise TypeError("Input image must be in 3-channel format!")

    if original_n_chan == 3:
        return rgb_img

    if normalize:
        if original_n_chan == 4:
            return rgb_to_rgba(rgb_img, alpha_ch)
        else:
            raise TypeError("No conversion for 3-channel to %d-channel images is implemented!" % original_n_chan)
    else:
        logger.warning("%s is converting %d-channel images to RGB images!" % (name, original_n_chan,))
        return rgb_img



from abc import ABC, abstractmethod


class LabelBehavior(ABC):
    """
    A LabelBehavior is an operation performed on the "true" label to
    """
    @abstractmethod
    def do(self, input_label: int) -> int:
        """
        Perform the actual desired label manipulation
        :param input_label: the input label to be manipulated
        :return: the manipulated label
        """
        pass



#!/usr/bin/env python3

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

__author__ = 'Kiran Karra, Chace Ashcraft, Nat Kavaler, Nathan Drenkow, Cash Costello'
__email__ = 'kiran.karra@jhuapl.edu,chace.ashcraft@jhuapl.edu,nathaniel.kavaler@jhaupl.edu,' \
            'nathan.drenkow@jhuapl.edu,cash.costello@jhuapl.edu'
__version__ = '0.2.0'


setuptools.setup(
    name='trojai',
    version=__version__,

    description='TrojAI model and dataset generation library',
    long_description=long_description,
    long_description_content_type="text/markdown",

    url = 'https://github.com/trojai/trojai',

    author=__author__,
    author_email=__email__,

    license='Apache License 2.0',

    python_requires='>=3',
    packages=['trojai', 'trojai.datagen', 'trojai.modelgen'],

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',

        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='deep-learning trojan adversarial',

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['numpy',
                      'pandas',
                      'scikit-image',
                      'joblib',
                      'scipy',
                      'pillow',
                      'scikit-learn==0.20.3',  # choose this version to be compatible with nosetests
                      'tqdm',
                      'opencv-python',
                      'torch',
                      'torchvision',
                      'torchtext',
                      ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
          'test': ['nose']
    },

    scripts=['scripts/datagen/mnist.py',
             'scripts/datagen/mnist_utils.py',
             'scripts/datagen/mnist_badnets.py',
             'scripts/datagen/mnist_badnets2.py',
             'scripts/modelgen/gen_and_train_mnist.py',
             'scripts/modelgen/gen_and_train_mnist_sequential.py'
             ],

    zip_safe=False
)


import logging

from numpy.random import RandomState
from trojai.datagen.conversion_utils import gray_to_rgb, rgba_to_rgb, rgb_to_rgba

from .entity import Entity, GenericEntity
from .transform import Transform

logger = logging.getLogger(__name__)

"""
Defines several transformations related to static (non-random) color manipulation
"""


class GrayscaleToRGBXForm(Transform):
    """ Converts an 3-channel grayscale image to RGB

    """
    def __init__(self) -> None:
        """
        Creates the object to perform the transformation.
        """
        pass

    def do(self, input_obj: Entity, random_state_obj: RandomState) -> Entity:
        """
        Convert the input object from 3-channel grayscale to RGB
        :param input_obj: Entity to be colorized
        :param random_state_obj: ignored
        :return: The colorized entity
        """
        img = input_obj.get_data()
        rgb_img = gray_to_rgb(img)
        logger.info("Converted input object from 3-channel grayscale to RGB")
        return GenericEntity(rgb_img, input_obj.get_mask())


class RGBAtoRGB(Transform):
    """ Converts input Entity from RGBA to RGB
    """
    def __init__(self) -> None:
        """
        Create the transformer object
        """
        pass

    def do(self, input_obj: Entity, random_state_obj: RandomState) -> Entity:
        """
        Perform the RGBA to RGB transformation
        :param input_obj: the Entity to be transformed
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        rgb_img, alpha_ch = rgba_to_rgb(img)
        logger.info("Converted input object from RGBA to RGB")
        return GenericEntity(rgb_img, input_obj.get_mask())


class RGBtoRGBA(Transform):
    """ Converts input Entity from RGB to RGBA

    """
    def __init__(self) -> None:
        """
        Create the transformer object
        """
        pass

    def do(self, input_obj: Entity, random_state_obj: RandomState) -> Entity:
        """
        Perform the RGBA to RGB transformation
        :param input_obj: the Entity to be transformed
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        rgba_img = rgb_to_rgba(img)
        logger.info("Converted input object from RGB to RGBA")
        return GenericEntity(rgba_img, input_obj.get_mask())



"""
This script downloads mnist data, creates and experiment with the alpha trigger from the Badnets paper
(https://arxiv.org/abs/1708.06733), and then trains a model with the architecture referenced in the same paper.

In this example, the model is trained on a 20% poisoned dataset for 300 epochs. Expected performance using pure
classification accuracy on clean and triggered data is ~99.2% on clean data and ~98.8% on triggered.
"""

import os
import argparse
from numpy.random import RandomState
import numpy as np
import logging.config

# some_file.py
import sys

# insert at 1, 0 is the script path (or '' in REPL)
sys.path.insert(1, os.path.abspath('../../trojai/scripts/datagen/'))

import mnist
from mnist_utils import download_and_extract_mnist_file, convert
import trojai.datagen.datatype_xforms as tdd
import trojai.datagen.insert_merges as tdi
import trojai.datagen.triggers as tdt
import trojai.datagen.common_behaviors as tdb
import trojai.datagen.experiment as tde
import trojai.datagen.config as tdc
import trojai.datagen.xform_merge_pipeline as tdx

import trojai.modelgen.data_manager as tpm_tdm
import trojai.modelgen.architecture_factory as tpm_af
import trojai.modelgen.architectures.mnist_architectures as tpma
import trojai.modelgen.config as tpmc
import trojai.modelgen.runner as tpmr
import trojai.modelgen.default_optimizer as tpm_do

import torch

import logging.config

logger = logging.getLogger(__name__)
MASTER_SEED = 1234


def download_mnist(clean_train_path, clean_test_path, temp_dir):
    # setup file system
    train_csv_dir = os.path.dirname(clean_train_path)
    test_csv_dir = os.path.dirname(clean_test_path)
    try:
        os.makedirs(train_csv_dir)
    except IOError:
        pass
    try:
        os.makedirs(test_csv_dir)
    except IOError:
        pass
    try:
        os.makedirs(temp_dir)
    except IOError:
        pass

    # download the 4 datasets
    logger.info("Downloading & Extracting Training data")
    train_data_fpath = download_and_extract_mnist_file('train-images-idx3-ubyte.gz', temp_dir)
    logger.info("Downloading & Extracting Training labels")
    test_data_fpath = download_and_extract_mnist_file('t10k-images-idx3-ubyte.gz', temp_dir)
    logger.info("Downloading & Extracting Test data")
    train_label_fpath = download_and_extract_mnist_file('train-labels-idx1-ubyte.gz', temp_dir)
    logger.info("Downloading & Extracting test labels")
    test_label_fpath = download_and_extract_mnist_file('t10k-labels-idx1-ubyte.gz', temp_dir)

    # convert it to the format we need
    logger.info("Converting Training data & Labels from ubyte to CSV")
    convert(train_data_fpath, train_label_fpath, clean_train_path, 60000, description='mnist_train_convert')
    logger.info("Converting Test data & Labels from ubyte to CSV")
    convert(test_data_fpath, test_label_fpath, clean_test_path, 10000, description='mnist_test_convert')

    logger.info("Cleaning up...")
    os.remove(os.path.join(temp_dir, 'train-images-idx3-ubyte.gz'))
    os.remove(os.path.join(temp_dir, 'train-labels-idx1-ubyte.gz'))
    os.remove(os.path.join(temp_dir, 't10k-images-idx3-ubyte.gz'))
    os.remove(os.path.join(temp_dir, 't10k-labels-idx1-ubyte.gz'))
    os.remove(os.path.join(temp_dir, 'train-images-idx3-ubyte'))
    os.remove(os.path.join(temp_dir, 'train-labels-idx1-ubyte'))
    os.remove(os.path.join(temp_dir, 't10k-images-idx3-ubyte'))
    os.remove(os.path.join(temp_dir, 't10k-labels-idx1-ubyte'))


def generate_mnist_experiment(train, test, output, train_output_csv_file, test_output_csv_file):
    logger.info("Generating experiment...")
    # Setup the files based on user inputs
    train_csv_file = os.path.abspath(train)
    test_csv_file = os.path.abspath(test)
    if not os.path.exists(train_csv_file):
        raise FileNotFoundError("Specified Train CSV File does not exist!")
    if not os.path.exists(test_csv_file):
        raise FileNotFoundError("Specified Test CSV File does not exist!")
    toplevel_folder = output

    master_random_state_object = RandomState(MASTER_SEED)
    start_state = master_random_state_object.get_state()

    # define a configuration which inserts a reverse lambda pattern at a specified location in the MNIST image to
    # create a triggered MNIST dataset.  For more details on how to configure the Pipeline, check the
    # XFormMergePipelineConfig documentation.  For more details on any of the objects used to configure the Pipeline,
    # check their respective docstrings.
    one_channel_alpha_trigger_cfg = \
        tdc.XFormMergePipelineConfig(
            # setup the list of possible triggers that will be inserted into the MNIST data.  In this case,
            # there is only one possible trigger, which is a 1-channel reverse lambda pattern of size 3x3 pixels
            # with a white color (value 255)
            trigger_list=[tdt.ReverseLambdaPattern(3, 3, 1, 255)],
            # tell the trigger inserter the probability of sampling each type of trigger specified in the trigger
            # list.  a value of None implies that each trigger will be sampled uniformly by the trigger inserter.
            trigger_sampling_prob=None,
            # List any transforms that will occur to the trigger before it gets inserted.  In this case, we do none.
            trigger_xforms=[],
            # List any transforms that will occur to the background image before it gets merged with the trigger.
            # Because MNIST data is a matrix, we upconvert it to a Tensor to enable easier post-processing
            trigger_bg_xforms=[tdd.ToTensorXForm()],
            # List how we merge the trigger and the background.  Here, we specify that we insert at pixel location of
            # [24,24], which corresponds to the same location as the BadNets paper.
            trigger_bg_merge=tdi.InsertAtLocation(np.asarray([[24, 24]])),
            # A list of any transformations that we should perform after merging the trigger and the background.
            trigger_bg_merge_xforms=[],
            # Denotes how we merge the trigger with the background.  In this case, we insert the trigger into the
            # image.  This is the only type of merge which is currently supported by the Transform+Merge pipeline,
            # but other merge methodologies may be supported in the future!
            merge_type='insert',
            # Specify that 15% of the clean data will be modified.  Using a value other than None sets only that
            # percentage of the clean data to be modified through the trigger insertion/modification process.
            per_class_trigger_frac=0.25
        )

    ############# Create the data ############
    # create the clean data
    clean_dataset_rootdir = os.path.join(toplevel_folder, 'mnist_clean')
    master_random_state_object.set_state(start_state)
    mnist.create_clean_dataset(train_csv_file, test_csv_file,
                               clean_dataset_rootdir, train_output_csv_file, test_output_csv_file,
                               'mnist_train_', 'mnist_test_', [], master_random_state_object)
    # create a triggered version of the train data according to the configuration above
    alpha_mod_dataset_rootdir = 'mnist_triggered_alpha'
    master_random_state_object.set_state(start_state)
    tdx.modify_clean_dataset(clean_dataset_rootdir, train_output_csv_file,
                             toplevel_folder, alpha_mod_dataset_rootdir,
                             one_channel_alpha_trigger_cfg, 'insert', master_random_state_object)
    # create a triggered version of the test data according to the configuration above
    master_random_state_object.set_state(start_state)
    tdx.modify_clean_dataset(clean_dataset_rootdir, test_output_csv_file,
                             toplevel_folder, alpha_mod_dataset_rootdir,
                             one_channel_alpha_trigger_cfg, 'insert', master_random_state_object)

    ############# Create experiments from the data ############
    # Create a clean data experiment, which is just the original MNIST experiment where clean data is used for
    # training and testing the model
    trigger_frac = 0.0
    trigger_behavior = tdb.WrappedAdd(1, 10)
    e = tde.ClassicExperiment(toplevel_folder, trigger_behavior)
    train_df = e.create_experiment(os.path.join(toplevel_folder, 'mnist_clean', 'train_mnist.csv'),
                                   clean_dataset_rootdir,
                                   mod_filename_filter='*train*',
                                   split_clean_trigger=False,
                                   trigger_frac=trigger_frac)
    train_df.to_csv(os.path.join(toplevel_folder, 'mnist_clean_experiment_train.csv'), index=None)
    test_clean_df, test_triggered_df = e.create_experiment(os.path.join(toplevel_folder, 'mnist_clean',
                                                                        'test_mnist.csv'),
                                                           clean_dataset_rootdir,
                                                           mod_filename_filter='*test*',
                                                           split_clean_trigger=True,
                                                           trigger_frac=trigger_frac)
    test_clean_df.to_csv(os.path.join(toplevel_folder, 'mnist_clean_experiment_test_clean.csv'), index=None)
    test_triggered_df.to_csv(os.path.join(toplevel_folder, 'mnist_clean_experiment_test_triggered.csv'), index=None)

    # Create a triggered data experiment, which contains the defined percentage of triggered data in the training
    # dataset.  The remaining training data is clean data.  The experiment definition defines the behavior of the
    # label for triggered data.  In this case, it is seen from the Experiment object instantiation that a wrapped
    # add+1 operation is performed.
    # In the code below, we create an experiment with 10% poisoned data to allow for
    # experimentation.
    trigger_frac = 0.2
    train_df = e.create_experiment(os.path.join(toplevel_folder, 'mnist_clean', 'train_mnist.csv'),
                                   os.path.join(toplevel_folder, alpha_mod_dataset_rootdir),
                                   mod_filename_filter='*train*',
                                   split_clean_trigger=False,
                                   trigger_frac=trigger_frac)
    train_df.to_csv(os.path.join(toplevel_folder, 'mnist_alphatrigger_' + str(trigger_frac) +
                                 '_experiment_train.csv'), index=None)
    test_clean_df, test_triggered_df = e.create_experiment(os.path.join(toplevel_folder,
                                                                        'mnist_clean', 'test_mnist.csv'),
                                                           os.path.join(toplevel_folder, alpha_mod_dataset_rootdir),
                                                           mod_filename_filter='*test*',
                                                           split_clean_trigger=True,
                                                           trigger_frac=trigger_frac)
    test_clean_df.to_csv(os.path.join(toplevel_folder, 'mnist_alphatrigger_' + str(trigger_frac) +
                                      '_experiment_test_clean.csv'), index=None)
    test_triggered_df.to_csv(os.path.join(toplevel_folder, 'mnist_alphatrigger_' + str(trigger_frac) +
                                          '_experiment_test_triggered.csv'), index=None)


def train_and_save_mnist_model(experiment_path, triggered_train, clean_test, triggered_test, model_save_dir,
                               parallel):
    logger.info("Training Model...")

    def img_transform(x):
        return x.unsqueeze(0)

    logging_params = {
        'num_batches_per_logmsg': 500,
        'tensorboard_output_dir': 'tensorboard_dir/',
        'experiment_name': 'badnets',
        'num_batches_per_metrics': 500,
        'num_batches_ver_val_dataset_metrics': None,
        'num_epochs_per_metric': 10
    }
    logging_cfg = tpmc.ReportingConfig(num_batches_per_logmsg=logging_params['num_batches_per_logmsg'],
                                       tensorboard_output_dir=logging_params['tensorboard_output_dir'],
                                       experiment_name=logging_params['experiment_name'],
                                       num_batches_per_metrics=logging_params['num_batches_per_metrics'],
                                       num_batches_ver_val_dataset_metrics=logging_params[
                                           'num_batches_ver_val_dataset_metrics'],
                                       num_epochs_per_metric=logging_params['num_epochs_per_metric'])

    # Train clean model to use as a base for triggered model
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    data_obj = tpm_tdm.DataManager(experiment_path,
                                   triggered_train,
                                   clean_test,
                                   triggered_test_file=triggered_test,
                                   data_transform=img_transform,
                                   shuffle_train=True)

    class MyArchFactory(tpm_af.ArchitectureFactory):
        def new_architecture(self):
            return tpma.BadNetExample()

    training_cfg = tpmc.TrainingConfig(device=device,
                                       epochs=300,
                                       batch_size=20,
                                       lr=1e-4)

    optim_cfg = tpmc.DefaultOptimizerConfig(training_cfg, logging_cfg)
    optim = tpm_do.DefaultOptimizer(optim_cfg)
    model_filename = 'BadNets_0.2_poison.pt'
    cfg = tpmc.RunnerConfig(MyArchFactory(), data_obj, optimizer=optim, model_save_dir=model_save_dir,
                            stats_save_dir=model_save_dir,
                            filename=model_filename,
                            parallel=parallel)
    runner = tpmr.Runner(cfg, {'script': 'gen_and_train_mnist.py'})
    runner.run()


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='MNIST Data Generation and Model Training Example')
    parser.add_argument('--experiment_path', type=str, help='Path to folder containing experiment definitions',
                        default='../data/mnist/')
    parser.add_argument('--train', type=str, help='CSV file which will contain MNIST Training data',
                        default='../data/mnist/clean/train.csv')
    parser.add_argument('--test', type=str, help='CSV file which will contain MNIST Test data',
                        default='../data/mnist/clean/test.csv')
    parser.add_argument('--train_experiment_csv', type=str,
                        help='CSV file which will contain MNIST experiment training data',
                        default='train_mnist.csv')
    parser.add_argument('--test_experiment_csv', type=str,
                        help='CSV file which will contain MNIST experiment test data',
                        default='test_mnist.csv')
    parser.add_argument('--log', type=str, help='Log File')
    parser.add_argument('--console', action='store_true')
    parser.add_argument('--models_output', type=str, default='BadNets_trained_models/',
                        help='Folder in which to save models')
    parser.add_argument('--tensorboard_dir', type=str, default='/tmp/tensorboard',
                        help='Folder for logging tensorboard')
    parser.add_argument('--gpu', action='store_true')
    parser.add_argument('--parallel', action='store_true', default=False,
                        help='Enable training with parallel processing, including multiple GPUs if available')
    a = parser.parse_args()

    use_gpu = False
    if a.gpu:
        # ensure it is available, otherwise revert to CPU training
        if torch.cuda.is_available():
            logger.info("Using GPU for training!")
            use_gpu = True
        else:
            logger.warning("Using CPU for training!")

    # setup logger
    handlers = []
    if a.log is not None:
        log_fname = a.log
        handlers.append('file')
    else:
        log_fname = '/dev/null'
    if a.console is not None:
        handlers.append('console')
    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'basic': {
                'format': '%(message)s',
            },
            'detailed': {
                'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
            },
        },
        'handlers': {
            'file': {
                'class': 'logging.handlers.RotatingFileHandler',
                'filename': log_fname,
                'maxBytes': 1 * 1024 * 1024,
                'backupCount': 5,
                'formatter': 'detailed',
                'level': 'INFO',
            },
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'basic',
                'level': 'INFO',
            }
        },
        'loggers': {
            'trojai_private': {  # todo: change to trojai after merging to trojai
                'handlers': handlers,
            },
        },
        'root': {
            'level': 'INFO',
        },
    })

    data_dir = a.experiment_path
    train = os.path.join(data_dir, 'clean/train.csv')
    test = os.path.join(data_dir, 'clean/test.csv')
    train_output_csv = a.train_experiment_csv
    test_output_csv = a.test_experiment_csv

    # Download mnist data if data directory doesn't exist
    # NOTE: This is not a full-proof way of making sure data exists! Make sure full data set is present or data_dir
    #   does not exist!
    if not os.path.isdir(data_dir):
        download_mnist(train, test, data_dir)

    # Generate triggered data and experiment files for mnist
    generate_mnist_experiment(train, test, data_dir, train_output_csv, test_output_csv)

    model_save_loc = os.path.join(data_dir, a.models_output, "mnist_alphatrigger_0.2/")

    # Train models using modelgen
    experiment_triggered_train = "mnist_alphatrigger_0.2_experiment_train.csv"
    experiment_clean_test = "mnist_alphatrigger_0.2_experiment_test_clean.csv"
    experiment_triggered_test = "mnist_alphatrigger_0.2_experiment_test_triggered.csv"

    train_and_save_mnist_model(data_dir, experiment_triggered_train, experiment_clean_test, experiment_triggered_test,
                               model_save_loc, a.parallel)
